/***************************************************************************
                          loadspace.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef _WIN32
#pragma warning (disable : 4786)
#endif

#include "loadspace.h"
#include "functions.h"
#include "errordlg.h"
#include "global.h"

#include <iterator>
#include <iostream>
#include <functional>

#include <map>

#ifdef _WIN32
#include "zlib.h"
#else
#include <zlib.h>
#endif

#include <qdom.h>
#include <qpoint.h>
#include <qcstring.h>
#include <qtextstream.h>

static map<QString,SpaceObjectPtr> objectStr;

void LoadSpace::createLoadData()
{
  objectStr.clear();
  objectStr["floor"] = new FloorObject( &objRef );
  objectStr["door"] = new DoorObject( &objRef );
  objectStr["entry"] = new EntryObject( &objRef );
  objectStr["bulkhead"] = new BulkHeadObject( &objRef );
  objectStr["exit"] = new ExitObject( &objRef );
  objectStr["extensible"] = new ExtensibleObject( &objRef );
}

LoadSpace::LoadSpace(const QString& fn, QWidget *p, ReferenceCenter* rc, SpaceMap& m, SpaceZones& z, SpaceDescription& d, ObjectRef& oref, ProgWnd& pr)
  : filename(fn)
  , parent(p)
  , refCenter(rc)
  , map(m)
  , zones(z)
  , desc(d)
  , prog( pr )
  , objRef( oref )
{ }

bool LoadSpace::operator()()
{
  createLoadData();
  QString buffer;
  gzFile f = gzopen(filename.latin1(),"rb");
  if(!f)
    {
      cerr << "Error when openinf file " << filename << endl;
      return false;
    }
  char car;
  while (gzread(f,&car,1)>0)
    {
      buffer+=car;
    }
  gzclose(f);

  QDomDocument doc;
  if(!doc.setContent(buffer))
    {
      cerr << "Error when loading file " << filename << endl;
      return false;
    }

  QDomElement level = doc.documentElement();
  // TODO : Here test the versions ...
  // TODO : Here test if the level is for spacehulk 1 or 2
  QDomElement genestealer_cfg, marine_cfg, global, xml_map;
  genestealer_cfg = level.lastChild().toElement();
  if((genestealer_cfg.isNull()) || (genestealer_cfg.tagName() != "genestealer_cfg"))
    {
      cerr << "genestealer_cfg is at the wrong position or not present at all"
	   << endl;
      return false;
    }
  marine_cfg = genestealer_cfg.previousSibling().toElement();
  if((marine_cfg.isNull()) || (marine_cfg.tagName() != "marine_cfg"))
    {
      cerr << "marine_cfg is at the wrong position or not present at all\n";
      return false;
    }
  global = marine_cfg.previousSibling().toElement();
  if((global.isNull()) || (global.tagName() != "global"))
    {
      cerr << "global is at the wrong position or not present at all\n";
      return false;
    }
  xml_map = global.previousSibling().toElement();
  if((xml_map.isNull()) || (xml_map.tagName() != "map"))
    {
      cerr << "Map is at the wrong position or not present at all\n";
      return false;
    }

  // First thing to do : set the map size ... 
  QString str_w, str_h;
  str_w = xml_map.attribute("width");
  str_h = xml_map.attribute("height");
  if(str_w.isNull())
    {
      cerr << "Error, the map must have a width" << endl;
      return false;
    }
  if(str_h.isNull())
    {
      cerr << "Error, the map must have a height" << endl;
      return false;
    }
  map.reset(str_w.toInt(), str_h.toInt());
  // Then load the zones
  try
    {
      loadZones(level);
      loadMap(xml_map);
      loadMarine( marine_cfg );
      loadGenestealer( genestealer_cfg );
      loadGlobal(global);
      zones.clean();
      refCenter->checkFailsafe();
    }
  catch(InvalidXmlException)
    {
      ErrorXmlSyntax dlg(parent);
      dlg.exec();
      return false;
    }
  catch(InvalidDataException)
    {
      ErrorBadData dlg(parent);
      dlg.exec();
      return false;
    }
  catch(InexistingRefException ie)
    {
      ErrorInexistingRef dlg(ie.name, parent);
      dlg.exec();
      return false;
    }
  catch(NameConflictException ne)
    {
      ErrorNameConflict dlg(ne.name, parent);
      dlg.exec();
      return false;
    }
  return true;
}

void LoadSpace::loadZones(QDomElement& root)
{
  QDomElement xml_zone = root.firstChild().toElement();
  while((!xml_zone.isNull()) && (xml_zone.tagName() == "zone"))
    {
      zones.setZone(xml_zone);
      xml_zone = xml_zone.nextSibling().toElement();
    }
  while((!xml_zone.isNull()) && (xml_zone.tagName() == "zone_set"))
    {
      zones.setZoneSet(xml_zone);
      xml_zone = xml_zone.nextSibling().toElement();
    }
}

void LoadSpace::loadMap(QDomElement& root)
{
  QDomElement xml_object = root.firstChild().toElement(), xml_zone;
  QString obj_name, zone_name;
  while(!xml_object.isNull())
    {
      obj_name = xml_object.tagName();
      SpaceObjectPtr obj = objectStr[obj_name]->getCopy();
      obj->setFromXml(xml_object);
      xml_zone = xml_object.firstChild().toElement();
      if(xml_zone.isNull())
	{
	  cerr << "Error, object " << obj_name << " need a zone as a child" << endl;
	  throw InvalidXmlException();
	}
      Referencer::reference_id zone_ref = zones.setZone(xml_zone);
      //Referencer::reference_id zone_ref = refCenter->getId( zone_name );
      set<int>& cases = zones.zones[ zone_ref ];
      set<int>::iterator it;
      for(it = cases.begin() ; it != cases.end() ; it++)
	{
	  map.getCase(*it)->insert(obj->getCopy());
	}
      // TODO : when the editor will be finished, I'll need to remove this !
      /*
      if(zone_name.contains("unnamed") && (!refCenter->referenced( zone_ref )))
	{
	  zones.eraseReference( zone_ref );
	}
        */
      xml_object = xml_object.nextSibling().toElement();
    }
}

void LoadSpace::loadGlobal(QDomElement& root)
{
  QDomElement xml_desc = root.firstChild().toElement();
  if(xml_desc.isNull())
    {
      cerr << "Error, no child to global ..." << endl;
      throw InvalidXmlException();
    }
  if(xml_desc.tagName() != "leveldesc")
    {
      cerr << "Error, first chulkd to global must be leveldesc and not " << xml_desc.tagName() << endl;
      throw InvalidXmlException();
    }
  desc.setFromXml(xml_desc);
  QDomElement next = xml_desc.nextSibling().toElement();
  if( next.isNull() )
    {
      cerr << "Error, sibling not an element ... " << endl;
      throw InvalidXmlException();
    }
  if( next.tagName() == "init" )
    {
      prog.setGlobalFromXml( next );
      next = next.nextSibling().toElement();
      if( next.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( next.tagName() == "draw" )
    {
      prog.setDrawFromXml( next );
      next = next.nextSibling().toElement();
      if( next.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( next.tagName() == "first" )
    {
      desc.setFromFirstXml( next );
      next = next.nextSibling().toElement();
      if( next.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  else
    {
      cerr << "Error, next sibling must be 'first'" << endl;
      throw InvalidXmlException();
    }
  if( next.tagName() == "winpriority" )
    {
      desc.setFromPriorXml( next );
    }
  else
    {
      cerr << "Error, next sibling must be 'winpriority'" << endl;
      throw InvalidXmlException();
    }
}

void LoadSpace::loadMarine( QDomElement& root )
{
  QDomElement child = root.firstChild().toElement();
  if( child.isNull() )
    {
      cerr << "Error, sibling not an element ... " << endl;
      throw InvalidXmlException();
    }
  if( child.tagName() == "init" )
    {
      prog.setMarineInitFromXml( child );
      child = child.nextSibling().toElement();
      if( child.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( child.tagName() == "begin" )
    {
      prog.setMarineBeginFromXml( child );
      child = child.nextSibling().toElement();
      if( child.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( child.tagName() == "win" )
    {
      prog.setMarineWinFromXml( child );
    }
  else
    {
      cerr << "Error, next sibling must be 'win'" << endl;
      throw InvalidXmlException();
    }

}

void LoadSpace::loadGenestealer( QDomElement& root )
{
  QDomElement child = root.firstChild().toElement();
  if( child.isNull() )
    {
      cerr << "Error, sibling not an element ... " << endl;
      throw InvalidXmlException();
    }
  if( child.tagName() == "init" )
    {
      prog.setGenestealerInitFromXml( child );
      child = child.nextSibling().toElement();
      if( child.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( child.tagName() == "begin" )
    {
      prog.setGenestealerBeginFromXml( child );
      child = child.nextSibling().toElement();
      if( child.isNull() )
        {
          cerr << "Error, sibling not an element ... " << endl;
          throw InvalidXmlException();
        }
    }
  if( child.tagName() == "win" )
    {
      prog.setGenestealerWinFromXml( child );
    }
  else
    {
      cerr << "Error, next sibling must be 'win'" << endl;
      throw InvalidXmlException();
    }
}

