/***************************************************************************
                          savespace.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "savespace.h"
#include "functions.h"
#include "global.h"

#include <iterator>
#include <iostream>
#include <functional>
#include <algorithm>

#include <map>

using namespace std;

#ifdef _WIN32
#include "zlib.h"
#else
#include <zlib.h>
#endif

#include <qdom.h>
#include <qpoint.h>
#include <qcstring.h>
#include <qtextstream.h>

struct saveObject
{
  saveObject(QDomElement& r, const SpaceMap& m, QDomDocument& d) : root(r), smap(m), doc(d) { }
  void operator()(const pair<SpaceObjectPtr,list<int> >& object)
    {
      QDomElement object_xml = object.first->getXmlElement(doc);
      QDomElement zone = doc.createElement("zone");
      for_each(object.second.begin(), object.second.end(), saveCase(zone,smap,doc));
      object_xml.appendChild(zone);
      root.appendChild(object_xml);
    }
protected:
  QDomElement& root;
  const SpaceMap& smap;
  QDomDocument& doc;
};

SaveSpace::SaveSpace(const QString& fn, const SpaceMap& m, SpaceZones& z, const SpaceDescription& d, const ProgWnd& p)
  : filename(fn)
  , smap(m)
  , zones(z)
  , desc(d)
  , prog( p )
{ }

bool SaveSpace::operator()()
{
  gzFile f = gzopen(filename,"wb8");
  if (!f) 
    {
      cout << "Cannot open file " << filename << endl;
      return false;
    }
  QDomDocument doc("level");
  QDomElement root = doc.createElement("level");
  root.setAttribute("version", 1.02);
  root.setAttribute("edition", "second");
  doc.appendChild(root);
  saveZones(root, doc);
  saveMap(root, doc);
  saveGlobal(root, doc);
  saveProg( root, doc );
  // Write XML doc on compressed file.
  QByteArray a;
  QTextStream ts(a, IO_WriteOnly);
  ts << doc.toString();
  char *toread = a.data();
  gzwrite(f,toread,a.size());
  gzclose(f);
  return true;
}

void SaveSpace::saveMap(QDomElement& root, QDomDocument& doc)
{
  map<SpaceObjectPtr,list<int>,PtrLess<SpaceObjectPtr> > objects;
  map<SpaceObjectPtr,list<int>,PtrLess<SpaceObjectPtr> >::iterator o_it;
  SpaceIterator it;
  for( it = smap.begin() ; it != smap.end() ; ++it )
    {
      objects[( *it ).second].push_back(( *it ).first);
    }
  QDomElement map_xml = doc.createElement("map");
  map_xml.setAttribute("width", smap.width());
  map_xml.setAttribute("height", smap.height());
  saveObject so( map_xml, smap, doc );
  cerr << "In object there is : " << endl;
  for( o_it = objects.begin() ; o_it != objects.end() ; ++o_it )
    {
      so( *o_it );
    }
  root.appendChild(map_xml);
}

void SaveSpace::saveZones(QDomElement& root, QDomDocument& doc)
{
  zones.setDoc(&doc);
  SpaceZones::zones_type::iterator it_z;
  for( it_z = zones.zones.begin() ; it_z != zones.zones.end() ; ++it_z )
    {
      root.appendChild( zones.getXmlElement( it_z->first ) );
    }
  SpaceZones::zone_sets_type::iterator it_zs;
  for( it_zs = zones.zone_sets.begin() ; it_zs != zones.zone_sets.end() ; ++it_zs )
    {
      root.appendChild( zones.getXmlElement( it_zs->first ) );
    }
}

void SaveSpace::saveGlobal(QDomElement& root, QDomDocument& doc)
{
  QDomElement global = doc.createElement("global");
  QDomElement xml_desc = desc.getXmlDescription(doc);
  global.appendChild(xml_desc);
  QDomElement global_init = prog.getXmlForGlobal( doc );
  QDomElement global_draw = prog.getXmlForDraw( doc );
  if( !global_init.isNull() )
    global.appendChild( global_init );
  if( !global_draw.isNull() )
    global.appendChild( global_draw );
  global.appendChild( desc.getXmlForFirst( doc ) );
  global.appendChild( desc.getXmlForPrior( doc ) );
  root.appendChild(global);
}

void SaveSpace::saveProg( QDomElement& root, QDomDocument& doc )
{
  QDomElement marine = doc.createElement( "marine_cfg" );
  QDomElement marine_init = prog.getXmlForMarineInit( doc );
  if( !marine_init.isNull() )
      marine.appendChild( marine_init );
  QDomElement marine_begin = prog.getXmlForMarineBegin( doc );
  if( !marine_begin.isNull() )
      marine.appendChild( marine_begin );
  QDomElement marine_win = prog.getXmlForMarineWin( doc );
  marine.appendChild( marine_win );
  root.appendChild( marine );
  QDomElement genestealer = doc.createElement( "genestealer_cfg" );
  QDomElement genestealer_init = prog.getXmlForGenestealerInit( doc );
  if( !genestealer_init.isNull() )
      genestealer.appendChild( genestealer_init );
  QDomElement genestealer_begin = prog.getXmlForGenestealerBegin( doc );
  if( !genestealer_begin.isNull() )
      genestealer.appendChild( genestealer_begin );
  QDomElement genestealer_win = prog.getXmlForGenestealerWin( doc );
  genestealer.appendChild( genestealer_win );
  root.appendChild( genestealer );
}

