/***************************************************************************
                          spacebars.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spacebars.h"
#include "global.h"
#include "functions.h"
#include "errordlg.h"

#include <qcombobox.h>
#include <qtoolbutton.h>
#include <qlabel.h>
#include <qinputdialog.h>

#include <iostream>

#include <algorithm>
#include <functional>

#if NEED_EXT
# include <ext/functional>
  using namespace __gnu_cxx;
#endif

using namespace std;

SpaceButton::SpaceButton(SpaceMapBar* b, SpaceToolPtr t, QToolBar * parent, const char * name)
  : QToolButton(t->getImage(), name, QString::null, NULL, NULL, parent, name)
  , tool(t)
  , bar(b)
{
  setToggleButton(true);
}

void SpaceButton::setOn ( bool enable )
{
  QToolButton::setOn(enable);
  pressed = enable;
  if(pressed)
    {
      action();
    }
  else
    {
      idle();
    }
}

void SpaceButton::toggle ()
{
  if(bar->map == NULL)
    {
      setOn(false);
    }
  QToolButton::toggle();
  if(pressed)
    {
      idle();
    }
  else
    {
      action();
    }
}

void SpaceButton::action()
{
  pressed = true;
  bar->changeButton(this);
  if(bar->map != NULL)
    {
      bar->map->setTool(tool);
    }
}

void SpaceButton::idle()
{
  pressed = false;
  bar->makeIdle(this);
  if(bar->map != NULL)
    {
      bar->map->setTool(NULL);
    }
}

SpaceMapBar::SpaceMapBar(const QString & label, QMainWindow *wnd)
  : QToolBar(label, wnd)
  , map(NULL)
  , last(NULL)
{
}

SpaceMapBar::~SpaceMapBar()
{ }

void SpaceMapBar::install(SpaceToolPtr tool)
{
  SpaceButton* button = new SpaceButton(this, tool, this, tool->getName().c_str());
  connect(button, SIGNAL(toggled(bool)), button, SLOT(setOn(bool)));
}

void SpaceMapBar::setMap(SpaceMap* m)
{
  map = m;
  if((map == NULL) && (last != NULL))
    {
      last->setOn(false);
    }
}

void SpaceMapBar::changeButton(SpaceButton* button)
{
  if(last != NULL)
    {
      last->setOn(false);
    }
  last = button;
  emit changeTool(button->textLabel());
}

void SpaceMapBar::makeIdle(const SpaceButton *button)
{
  if(last == button)
    {
      last = NULL;
    }
  emit changeTool("No tool");
}

SpaceZonesBar::SpaceZonesBar(const QString & label, QMainWindow *wnd)
  : QToolBar(label, wnd)
  , zones(NULL)
  , map(NULL)
  , unshow_sel( NULL )
  , zones_label( NULL )
  , set_label( NULL )
  , cb_zones( NULL )
  , validate( NULL )
  , rename_zone( NULL )
  , erase_zone( NULL )
  , cb_zonesets( NULL )
  , edit_set( NULL )
  , erase_set( NULL )
{
  static QPixmap unselect(getEditorDir() + "unselect.png");
  unshow_sel = new QToolButton(unselect, "Unselect all","", this, SLOT(unshowZoneSel()), this);
  zones_label = new QLabel("Zones: ", this);
  cb_zones = new QComboBox(true, this, "zones list");
  connect(cb_zones, SIGNAL(activated ( const QString &)), this, SLOT(showZone(const QString&)));
  static QPixmap valid_img(getEditorDir() + "validate.png");
  validate = new QToolButton(valid_img, "Validate zone", "", this, SLOT(setZone()), this);
  static QPixmap rename_img( getEditorDir() + "rename.png" );
  rename_zone = new QToolButton( rename_img, "Rename zone", "", this, SLOT( renameZone() ), this );
  static QPixmap erase_img(getEditorDir() + "delete.png");
  erase_zone = new QToolButton(erase_img, "Delete zone", "", this, SLOT(deleteZone()), this);
  addSeparator();
  set_label = new QLabel("Zone sets: ", this);
  cb_zonesets = new QComboBox(true, this, "zone sets list");
  connect(cb_zonesets, SIGNAL(activated(const QString &)), this, SLOT(showZoneSet(const QString&)));
  static QPixmap edit_img(getEditorDir() + "edit.png");
  edit_set = new QToolButton(edit_img, "Edit zone set", "", this, SLOT(editZoneSet()), this); 
  erase_set = new QToolButton(erase_img, "Delete zone set", "", this, SLOT(deleteZoneSet()), this);
}

SpaceZonesBar::~SpaceZonesBar()
{ }

void SpaceZonesBar::setZones(SpaceZones* z, SpaceMap* m)
{
  zones = z;
  map = m;
}

void SpaceZonesBar::setZone()
{
  if(zones != NULL)
    {
      QString name = cb_zones->currentText();
      if(!name.isEmpty())
        {
          try
            {
              zones->setZone(name);
            }
          catch(InvalidDataException)
            {
              ErrorSetZone dlg(this);
              dlg.exec();
              return;
            }
          catch(NameConflictException ne)
            {
              ErrorNameConflict dlg(ne.name, this);
              dlg.exec();
              return;
            }
          reloadZones();
        }
    }
}

void SpaceZonesBar::renameZone()
{
  QString name = cb_zones->currentText();
  if( !name.isEmpty() )
    {
      try
        {
          Referencer::reference_id ref = zones->getId( name );
          QString new_name = QInputDialog::getText( "Rename the zone " + name, "New name : ",
                                                    QLineEdit::Normal, name );
          if( !new_name.isEmpty() )
            {
              zones->renameRef( ref, new_name );
              reloadZones();
            }
        }
      catch( NameConflictException ne )
        {
          ErrorNameConflict dlg( ne.name, this );
          dlg.exec();
        }
      catch( InexistingRefException ie )
        {
          ErrorInexistingRef dlg( ie.name, this );
          dlg.exec();
        }
    }
}

void SpaceZonesBar::deleteZone()
{
  if(zones != NULL)
    {
      QString name = cb_zones->currentText();
      if(!name.isEmpty())
        {
          try
            {
              Referencer::reference_id ref = zones->getId( name );
              zones->eraseReference( ref );
            }
          catch(InexistingRefException ie)
            {
              ErrorInexistingRef dlg(ie.name, this);
              dlg.exec();
              return;
            }
          catch(StillRefException se)
            {
              ErrorStillRef dlg(se.name, this);
              dlg.exec();
              return;
            }
          unshowZoneSel();
          reloadZones();
        }
    }
}

void SpaceZonesBar::editZoneSet()
{
  if(zones != NULL)
    {
      QString name = cb_zonesets->currentText();
      try
        {
          zones->setZoneSet( name );
        }
      catch(InexistingRefException ie)
        {
          ErrorInexistingRef dlg(ie.name, this);
          dlg.exec();
          return;
        }
      catch(InvalidDataException)
        {
          ErrorZoneSet dlg(this);
          dlg.exec();
          return;
        }
      catch(NameConflictException ne)
        {
          ErrorNameConflict dlg(ne.name, this);
          dlg.exec();
          return;
        }
      reloadSets();
    }
}

void SpaceZonesBar::deleteZoneSet()
{
  if(zones != NULL)
    {
      QString name = cb_zonesets->currentText();
      if(!name.isEmpty())
        {
          try
            {
              Referencer::reference_id ref = zones->getId( name );
              zones->eraseReference( ref );
            }
          catch(InexistingRefException ie)
            {
              ErrorInexistingRef dlg(ie.name, this);
              dlg.exec();
              return;
            }
          catch(StillRefException se)
            {
              ErrorStillRef dlg(se.name, this);
              dlg.exec();
              return;
            }
          unshowZoneSel();
          reloadSets();
        }
    }
}

void SpaceZonesBar::reloadZones()
{
  cb_zones->clear();
  SpaceZones::zones_type::iterator it_z;
  for( it_z = zones->zones.begin() ; it_z != zones->zones.end() ; ++it_z )
    {
      cb_zones->insertItem( zones->getName( it_z->first ) );
    }
}

void SpaceZonesBar::reloadSets()
{
  cb_zonesets->clear();
  SpaceZones::zone_sets_type::iterator it_zs;
  for( it_zs = zones->zone_sets.begin() ; it_zs != zones->zone_sets.end() ; ++it_zs )
    {
      cb_zonesets->insertItem( zones->getName( it_zs->first ) );
    }
}

void SpaceZonesBar::showZone(const QString& name)
{
  if(zones != NULL)
    {
      try
        {
          Referencer::reference_id ref = zones->getId( name );
          SpaceZones::zones_type::iterator it = zones->zones.find( ref );
          if(it != zones->zones.end())
            {
              unshowZoneSel();
              set<int>::iterator it_qs;
              for( it_qs = it->second.begin() ; it_qs != it->second.end() ; ++it_qs )
                {
                  map->getCase( *it_qs )->toggle_select();
                }
            }
        }
      catch( InexistingRefException ie )
        {
          ErrorInexistingRef dlg( ie.name );
          dlg.exec();
        }
    }
}

void SpaceZonesBar::showZoneSet(const QString& name)
{
  if(zones != NULL)
    {
      try
        {
          Referencer::reference_id ref = zones->getId( name );
          SpaceZones::zone_sets_type::iterator it = zones->zone_sets.find( ref );
          if(it != zones->zone_sets.end())
            {
              unshowZoneSel();
              set<Referencer::reference_id>::iterator z_it;
              for(z_it = it->second.begin() ; z_it != it->second.end() ; z_it++)
                {
                  set<int>& cases = zones->zones[*z_it];
                  set<int>::iterator it_c;
                  for( it_c = cases.begin() ; it_c != cases.end() ; ++it_c )
                    {
                      map->getCase( *it_c )->select( true );
                    }
                }
            }
        }
      catch( InexistingRefException ie )
        {
          ErrorInexistingRef dlg( ie.name );
          dlg.exec();
        }
    }
}

void SpaceZonesBar::reload()
{
  reloadZones();
  reloadSets();
}

void SpaceZonesBar::unshowZoneSel()
{
  SpaceMap::case_iterator it_case;
  for( it_case = map->case_begin() ; it_case != map->case_end() ; ++it_case )
    {
      map->getCase( it_case->first )->select( false );
    }
}

