/***************************************************************************
                          spacedesc.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spacedesc.h"
#include "levelinfo.h"
#include "spacezones.h"
#include "functions.h"
#include "spaceref.h"
#include "global.h"

#include <qlineedit.h>
#include <qmultilineedit.h>
#include <qcombobox.h>
#include <qlistbox.h>
#include <algorithm>

#include <iostream>

using namespace std;

SpaceDescription::SpaceDescription(QWidget * p, SpaceZones* z, ReferenceCenter* rc)
  : parent(p)
  , zones(z)
  , refCenter(rc)
  , title("Mission title")
  , author("")
  , narration("")
  , objectives("")
  , forces("")
  , conditions("")
  , objectiveZone( Referencer::null_reference )
  , win_prior( "marine" )
  , first_player( "marine" )
{ }

void SpaceDescription::edit()
{
  LevelInfo dlg(parent, "", true);
  dlg.title->setText(title);
  dlg.author->setText(author);
  dlg.narration->setText(narration);
  dlg.objectives->setText(objectives);
  dlg.forces->setText(forces);
  dlg.conditions->setText(conditions);
  dlg.objectiveZone->insertItem("None");
  SpaceZones::zone_sets_type::iterator it_zs;
  for( it_zs = zones->zone_sets.begin() ; it_zs != zones->zone_sets.end() ; ++it_zs )
    {
      dlg.objectiveZone->insertItem( refCenter->getName( it_zs->first ) );
    }
  QListBox *lb = dlg.objectiveZone->listBox();
  for(size_t i = 0 ; i < lb->count() ; i++)
    {
      if(lb->text(i) == objectiveZone)
	{
	  dlg.objectiveZone->setCurrentItem(i);
	  break;
	}
    }
  if( first_player == "marine" )
    {
      dlg.firstToPlay->setCurrentItem( 0 );
    }
  else
    {
      dlg.firstToPlay->setCurrentItem( 1 );
    }
  if( win_prior == "marine" )
    {
      dlg.winPriority->setCurrentItem( 0 );
    }
  else
    {
      dlg.winPriority->setCurrentItem( 1 );
    }
  if(dlg.exec() == QDialog::Accepted)
    {
      title = dlg.title->text();
      author = dlg.author->text();
      narration = dlg.narration->text();
      objectives = dlg.objectives->text();
      forces = dlg.forces->text();
      conditions = dlg.conditions->text();
      QString obj_z = dlg.objectiveZone->currentText();
      if( obj_z == "None" )
        {
          if( objectiveZone != Referencer::null_reference )
            {
              refCenter->unbind( objectiveZone );
              objectiveZone = Referencer::null_reference;
            }
        }
      else
        {
          Referencer::reference_id newref = refCenter->getId( obj_z );
          if( newref != objectiveZone)
	    {
	      if(objectiveZone != Referencer::null_reference )
	        {
	          refCenter->unbind(objectiveZone);
	        }
	      refCenter->bind(newref);
	    }
          objectiveZone = newref;
        }
      first_player = dlg.firstToPlay->currentText().lower();
      win_prior = dlg.winPriority->currentText().lower();
    }
}

void SpaceDescription::setFromXml(QDomElement& root)
{
  QDomElement xml_field = root.firstChild().toElement();
  if(xml_field.tagName() != "title")
    {
      cerr << "Error, first descriptor should be the title" << endl;
      throw InvalidXmlException();
    }
  QDomText xml_title = xml_field.firstChild().toText();
  if(xml_title.isNull())
    {
      cerr << "Title must contain text" << endl;
      throw InvalidXmlException();
    }
  title = xml_title.data();
  xml_field = xml_field.nextSibling().toElement();
  map<QString, QString*> fields;
  map<QString, QString*>::iterator it;
  fields["author"] = &author;
  fields["narration"] = &narration;
  fields["objectives"] = &objectives;
  fields["forces"] = &forces;
  fields["conditions"] = &conditions;
  QDomText xml_data;
  while(!xml_field.isNull())
    {
      if(xml_field.tagName() == "objective_squares")
	{
	  objectiveZone = zones->setZoneSet(xml_field.firstChild().toElement());
	  refCenter->bind(objectiveZone);
	}
      else
	{
	  it = fields.find(xml_field.tagName());
	  if(it == fields.end())
	    {
	      cerr << "Unknown field : " << xml_field.tagName() << endl;
	      throw InvalidXmlException();
	    }
	  xml_data = xml_field.firstChild().toText();
	  if(xml_title.isNull())
	    {
	      cerr << "Field " << xml_field.tagName() << " must contain text" << endl;
	      throw InvalidXmlException();
	    }
	  *(it->second) = xml_data.data();
	}
      xml_field = xml_field.nextSibling().toElement();
    }
}

void SpaceDescription::setFromFirstXml(QDomElement& root)
{
  if( !root.hasAttribute( "player" ) )
    {
      cerr << "first tag must contains 'player' attribute" << endl;
      throw InvalidXmlException();
    }
  first_player = root.attribute( "player" );
}

void SpaceDescription::setFromPriorXml(QDomElement& root)
{
  if( !root.hasAttribute( "player" ) )
    {
      cerr << "winpriority tag must contains 'player' attribute" << endl;
      throw InvalidXmlException();
    }
  win_prior = root.attribute( "player" );
}

QDomElement SpaceDescription::getXmlDescription(QDomDocument& doc) const
{
  QDomElement result = doc.createElement("leveldesc");
  QDomElement field;
  QDomText text;
  list<pair<QString, QString> > fields;
  fields.push_back(make_pair(QString( "title" ), title));
  if(author != "")
    {
      fields.push_back(make_pair(QString( "author" ), author));
    }
  if(narration != "")
    {
      fields.push_back(make_pair(QString( "narration" ), narration));
    }
  if(objectives != "")
    {
      fields.push_back(make_pair(QString( "objectives" ), objectives));
    }
  if(forces != "")
    {
      fields.push_back(make_pair(QString( "forces" ), forces));
    }
  if(conditions != "")
    {
      fields.push_back(make_pair(QString( "conditions" ), conditions));
    }
  list<pair<QString, QString> >::iterator it;
  for(it = fields.begin() ; it != fields.end() ; it++)
    {
      field = doc.createElement(it->first);
      text = doc.createTextNode(it->second);
      field.appendChild(text);
      result.appendChild(field);
    }
  if(objectiveZone != Referencer::null_reference)
    {
      QDomElement objective_squares = doc.createElement("objective_squares");
      saveRef( "zone_set", objective_squares, doc )( refCenter->getName( objectiveZone ) );
      result.appendChild(objective_squares);
    }
  return result;
}

QDomElement SpaceDescription::getXmlForFirst(QDomDocument& doc) const
{
  QDomElement result = doc.createElement("first");
  result.setAttribute( "player", first_player );
  return result;
}

QDomElement SpaceDescription::getXmlForPrior(QDomDocument& doc) const
{
  QDomElement result = doc.createElement("winpriority");
  result.setAttribute( "player", win_prior );
  return result;
}

void SpaceDescription::clear()
{
  title = "Mission title";
  author = "";
  narration = "";
  objectives = "";
  forces = "";
  conditions = "";
  objectiveZone = Referencer::null_reference;
  first_player = "Marine";
  win_prior = "Marine";
}

