/***************************************************************************
                          spaceditor.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spaceditor.h"
#include "newmapdialog.h"
#include "spacetools.h"
#include "spacebars.h"
#include "savespace.h"
#include "spacezones.h"
#include "loadspace.h"
#include "spacedesc.h"
#include "progeditor.h"
#include "objectref.h"

#include <iostream>

#include <qspinbox.h>
#include <qfiledialog.h>
#include <qpopupmenu.h>
#include <qscrollview.h>
#include <qstatusbar.h>
#include <qmenubar.h>
#include <qlabel.h>
#include <qapplication.h>
#include <qdir.h>

#include <algorithm>
#include <functional>

#if NEED_EXT
# include <ext/functional>
  using namespace __gnu_cxx;
#endif

using namespace std;

#ifdef _WIN32
//#  error Have to implement homeDir and spaceDir
#include "Windows.h"
#define DIR_DELIMITER '\\'
QString homeDir=".";
QString spaceDir=".";
void GetDirFromRegistry(void);
QString getEditorDir()
{
    return spaceDir + "\\editor\\";
}
QString getThemeDir()
{
    return homeDir + "\\themes\\Deathwing\\";
}
#else
#define DIR_DELIMITER '/'
QString homeDir = QDir::homeDirPath()+"/.spacehulk";
QString spaceDir = DATA_DIR;
QString getEditorDir()
{
    return spaceDir + "/editor/";
}
QString getThemeDir()
{
    return spaceDir + "/themes/Deathwing/";
}

#endif

#ifdef _WIN32
int WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, char* lpCmdLine, int nShowCmd)
{
  GetDirFromRegistry();
  char* cmdline = GetCommandLine();
  char **argv = NULL;
  int argc;
  int i;
  LPCWSTR unicode_cmdline = GetCommandLineW();
  LPWSTR *unicode_argv = CommandLineToArgvW( unicode_cmdline, &argc);
  argv = new char*[argc];
  argv[0] = cmdline;
  for(i = 0 ; i < argc ; ++i)
    {
      int len = lstrlenW( unicode_argv[0]);
      *(argv[i] + len) = '\0';
      if( i+1 < argc )
        {
          argv[i+1] = argv[i] + len + 1;
        }
    }
  SpacEditor edit(argc, argv);
  return edit.exec();
}
#else
int main(int argc, char *argv[])
{
  SpacEditor edit(argc, argv);
  return edit.exec();
}
#endif

SpacEditor::SpacEditor(int argc, char *argv[])
  : QApplication(argc, argv)
  , menu( NULL )
  , status( NULL )
  , fileMenu(NULL)
  , editMenu(NULL)
  , helpMenu(NULL)
  , viewMenu( NULL )
  , smap(NULL)
  , zones(NULL)
  , refCenter(NULL)
  , desc(NULL)
  , prog( NULL )
  , objRef( NULL )
  , mainBox( NULL )
  , mainView( NULL )
  , mapBar(NULL)
  , zonesBar(NULL)
  , statusBar( NULL )
  , currentTool( NULL )
{
  menu = main_wnd.menuBar();
  createMenu();
  status = main_wnd.statusBar();
  createStatusBar();
  mainView = new QScrollView(&main_wnd);
  main_wnd.setCentralWidget(mainView);
  mainBox = mainView->viewport();
  mainView->show();
  mapBar = new SpaceMapBar("Map tools", &main_wnd);
  mapBar->setLabel( tr("Map tools"));
  connect(mapBar, SIGNAL(changeTool(const QString&)), this, SLOT(setTool(const QString&)));
  RegisterTools();
  main_wnd.addToolBar(mapBar);
  mapBar->show();
  zonesBar = new SpaceZonesBar("Zones edition", &main_wnd);
  zonesBar->setLabel( tr("Zones edition"));
  main_wnd.addToolBar(zonesBar);
  zonesBar->show();
  main_wnd.show();
  setMainWidget(&main_wnd);
}

SpacEditor::~SpacEditor()
{
}

void SpacEditor::createMenu()
{
  fileMenu = new QPopupMenu;
  fileMenu->insertItem("&New", this, SLOT(newFile()), CTRL+Key_N );
  fileMenu->insertItem("&Open", this, SLOT(openFile()), CTRL+Key_O );
  fileMenu->insertItem("&Save", this, SLOT(saveFile()), CTRL+Key_S );
  fileMenu->insertItem("&Close", this, SLOT(closeFile()), CTRL+Key_W );
  fileMenu->insertSeparator();
  fileMenu->insertItem("&Quit", this, SLOT(quit()), CTRL+Key_Q);
  menu->insertItem("&File", fileMenu);

  editMenu = new QPopupMenu;
  editMenu->insertItem("&Unselect all", this, SLOT(unselectEdit()), CTRL+Key_U);
  editMenu->insertItem("&Select all", this, SLOT(selectAllEdit()), CTRL+Key_A);
  editMenu->insertItem("&Inverse selection", this, SLOT(inverseSelectEdit()), CTRL+Key_I);
  editMenu->insertSeparator();
  editMenu->insertItem("Edit description", this, SLOT(editDesc()), CTRL+Key_D);
  menu->insertItem("&Edit", editMenu);

  viewMenu = new QPopupMenu;
  changeSizeId = 18642;
  viewMenu->insertItem("S&mall view", this, SLOT(changeSize()), CTRL+Key_M, changeSizeId);

  viewMenu->setCheckable(true);
  connect(viewMenu, SIGNAL(activated(int)), this, SLOT(setCheck(int)));
  menu->insertItem("&View", viewMenu);

  helpMenu = new QPopupMenu;
}

void SpacEditor::createStatusBar()
{
  statusBar = main_wnd.statusBar();
  fileLoaded = new QLabel(statusBar);
  fileLoaded->setText("No file loaded");
  fileLoaded->setMinimumWidth(100);
  currentTool = new QLabel(statusBar);
  currentTool->setText("No tool selected");
  currentTool->setMinimumWidth(100);
  statusBar->addWidget(currentTool);
  statusBar->addWidget(fileLoaded);
}

void SpacEditor::newFile()
{
  cout << "New file" << endl;
  NewMapDialog dlg(&main_wnd, "New map dialog", TRUE);
  int value = dlg.exec();
  if(value == QDialog::Accepted)
    {
      int width, height;
      width = dlg.mapWidth->text().toInt();
      height = dlg.mapHeight->text().toInt();
      cout << "Width = " << width << "\nHeight = " << height << endl;
      if( smap != NULL )
        {
          resetObjects( width, height );
        }
      else
        {
          createObjects(width, height);
        }
    }
  else
    {
      cout << "Cancelled" << endl;
    }
}

void SpacEditor::openFile()
{
  cout << "Open file" << endl;
  QString filename = QFileDialog::getOpenFileName(spaceDir+DIR_DELIMITER+"levels", "Levels (*.lvl)", &main_wnd);
  if(!filename.isEmpty())
    {
      cout << "Selected file : " << filename << endl;
      if(smap == NULL)
        {
          createObjects(1,1);
        }
      else
        {
          resetObjects(1,1);
        }
      LoadSpace load(filename, &main_wnd, refCenter, *smap, *zones, *desc, *objRef, *prog->getProgWnd() );
      if(!load())
        {
          deleteObjects();
          return;
        }
      zonesBar->reload();
      setFile(filename);
    }
  else
    {
      cout << "Aborted" << endl;
    }
}

void SpacEditor::saveFile()
{
  QString filename = QFileDialog::getSaveFileName(homeDir + DIR_DELIMITER + "levels", "Levels (*.lvl)", &main_wnd);
  if(!filename.isEmpty())
    {
      cout << "Selected file : " << filename << endl;
      SaveSpace(filename, *smap, *zones, *desc, *prog->getProgWnd() )();
      setFile(filename);
    }
  else
    {
      cout << "Aborted" << endl;
    }
}

void SpacEditor::closeFile()
{
  cout << "Close file" << endl;
  if(smap != NULL)
    {
      deleteObjects();
      setFile("No file loaded");
    }
}

void SpacEditor::quit()
{
  cout << "Quit" << endl;
  QApplication::exit(0);
}

void SpacEditor::RegisterTools()
{
  mapBar->install(new FloorTool);
  mapBar->install(new DoorTool);
  mapBar->install(new BulkHeadTool);
  mapBar->install(new EntryTool);
  mapBar->install(new ExitTool);
  mapBar->install(new ExtensibleTool);
  mapBar->install(new SelectTool);
  mapBar->install(new EraseTool);
  mapBar->install(new DeleteOneTool);
}

void SpacEditor::selectAllEdit()
{
  if(smap != NULL)
    {
      for(int i = 0 ; i < smap->width()*smap->height(); i++)
        {
          smap->getCase(i)->select(true);
        }
    }
}

void SpacEditor::inverseSelectEdit()
{
  if(smap != NULL)
    {
      for(int i = 0 ; i < smap->width()*smap->height(); i++)
        {
          smap->getCase(i)->toggle_select();
        }
    }
}

void SpacEditor::unselectEdit()
{
  if(smap != NULL)
    {
      zonesBar->unshowZoneSel();
    }
}

void SpacEditor::createObjects(int w, int h)
{
  refCenter = new ReferenceCenter;
  objRef = new ObjectRef( refCenter );
  smap = new SpaceMap(w,h,mainBox);
  smap->setRef( objRef );
  mainView->addChild(smap);
  smap->show();
  mapBar->setMap(smap);
  zones = new SpaceZones(refCenter);
  zones->setMap(smap);
  zonesBar->setZones(zones, smap);
  desc = new SpaceDescription(smap, zones, refCenter);
  prog = new ProgEditor;
  prog->setRef( smap, zones, refCenter, objRef );
  prog->show();
}

void SpacEditor::resetObjects(int w, int h)
{
  objRef->clear();
  smap->reset(w,h);
  refCenter->clear();
  zones->clear();
  desc->clear();
  delete prog;
  prog = new ProgEditor;
  prog->setRef( smap, zones, refCenter, objRef );
  prog->show();
}

void SpacEditor::deleteObjects()
{
  delete prog;
  delete desc;
  delete smap;
  delete zones;
  delete refCenter;
  smap = NULL;
  zones = NULL;
  refCenter = NULL;
  desc = NULL;
  mapBar->setMap(NULL);
  zonesBar->setZones(NULL,NULL);
  prog = NULL;
}

void SpacEditor::setTool(const QString& toolname)
{
  currentTool->setText(toolname);
}

void SpacEditor::setFile(const QString& filename)
{
  fileLoaded->setText(filename);
}

void SpacEditor::editDesc()
{
  if(desc != NULL)
    {
      desc->edit();
    }
}

void SpacEditor::setCheck(int itemId)
{
  if(smap != NULL)
    {
      if(viewMenu->isItemChecked(itemId))
        {
          viewMenu->setItemChecked(itemId, false);
        }
      else
        {
          viewMenu->setItemChecked(itemId, true);
        }
    }
}

void SpacEditor::changeSize()
{
  if(smap != NULL)
    {
      smap->setSmall(viewMenu->isItemChecked(changeSizeId));
    }
}


#ifdef _WIN32
#include <Windows.h>
#include <Winreg.h>

void GetDirFromRegistry()
{
  HKEY hkey;
  long ret;
  char home[1024];
  DWORD size=1024;

  ret = RegOpenKeyEx(HKEY_CLASSES_ROOT,"QSpaceHulk",0,KEY_QUERY_VALUE, &hkey);
  if (ret != ERROR_SUCCESS)
    return;
  ret = RegQueryValueEx(hkey, "HomeDir", NULL, NULL, (LPBYTE) home, &size);
  if (ret != ERROR_SUCCESS)
    return;
  homeDir = home;
  size = 1024;
  ret = RegQueryValueEx(hkey, "EditorDir", NULL, NULL, (LPBYTE) home, &size);
  if (ret != ERROR_SUCCESS)
    return;
  spaceDir = home;
  RegCloseKey(hkey);
}
#endif


