/***************************************************************************
                          spacemap.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spacemap.h"
#include "spacetools.h"

#include <qpen.h>
#include <qpainter.h>
#include <qpopupmenu.h>

#include <algorithm>
#include <iostream>
#include <functional>

#if NEED_EXT
# include <ext/functional>
  using namespace __gnu_cxx;
#endif
#include <iterator>

using namespace std;

#define UNSELECT QColor(128,128,128)
#define ONMOUSE QColor(255,255,0)
#define SELECTED QColor(255,0,0)
#define BACKGROUND QColor(0,0,0)

bool LevelCmp::operator()(const SpaceObjectPtr& p1, const SpaceObjectPtr& p2) const
{
  return p1->getLevel() < p2->getLevel();
}

bool LevelEqual::operator()(const SpaceObjectPtr& p1, const SpaceObjectPtr& p2) const
{
  return p1->getLevel() == p2->getLevel();
}

static QPixmap *selected_big = NULL;
static QPixmap *selected_small = NULL;

SpaceCase::SpaceCase(int p, SpaceMap *parrent)
  : m_selected( false )
  , pos( p )
  , smap( parrent )
{
  if( selected_big == NULL )
    {
      selected_big = new QPixmap( getEditorDir() + "select_big.png" );
    }
  if( selected_small == NULL )
    {
      selected_small = new QPixmap( getEditorDir() + "select_small.png" );
    }
}

void SpaceCase::toggle_select()
{
  m_selected ^= true;
  repaint();
}

void SpaceCase::select(bool s)
{
  m_selected = s;
  repaint();
}

pair<SpaceCase::iterator,bool> SpaceCase::insert(SpaceObjectPtr object)
{
  pair<iterator,bool> result = set<SpaceObjectPtr,LevelCmp>::insert(object);
  if(!result.second)
    {
      erase(find_if(begin(), end(),
		    bind2nd(LevelEqual(), object)));
      set<SpaceObjectPtr,LevelCmp>::insert(object);
    }
  repaint();
  return result;
}

void SpaceCase::repaint()
{
  QRect rect = smap->caseRect(pos);
  QPainter paint(smap->getPixmap());
  paint.fillRect(rect, BACKGROUND);
  paint.setPen(UNSELECT);
  paint.drawRect(rect);
  QPoint pt = rect.topLeft()+QPoint(1,1);
  iterator it;
  QPixmap* selection;
  if( smap->isSmall )
    {
      selection = selected_small;
    }
  else
    {
      selection = selected_big;
    }
  for(it = begin() ; it != end() ; it++)
    {
      paint.drawPixmap(pt, (*it)->getImage(smap->isSmall));
    }
  if(m_selected)
    {
      //paint.fillRect(rect, QBrush(SELECTED, QBrush::Dense6Pattern));
      paint.drawPixmap( pt, *selection );
    }
  smap->update(rect);
}

void SpaceCase::deleteItem(int pos)
{
  iterator it;
  int n;
  for(n = 0, it = begin() ; it != end() ; it++, n++)
    {
      if(n == pos)
	{
	  erase(it);
	  break;
	}
    }
  repaint();
}

bool SpaceCase::selected()
{
  return m_selected;
}

SpaceIterator::SpaceIterator()
{ }

SpaceIterator::SpaceIterator(const map<int,SpaceCasePtr>& smap, bool begin)
{
  if(begin)
    {
      end = smap.end();
      main_it = smap.begin();
      if( main_it == smap.end() )
        return;
      local_it = main_it->second->begin();
      while(local_it == main_it->second->end())
	{
	  if((++main_it) == end)
	    {
	      return;
	    }
	  local_it = main_it->second->begin();
	}
    }
  else
    {
      end = main_it = smap.end();
    }
}

SpaceIterator::SpaceIterator(const SpaceIterator& copy)
  : main_it(copy.main_it)
  , end(end)
  , local_it(copy.local_it)
{
}

pair<int, SpaceObjectPtr> SpaceIterator::operator*()
{
  return pair<int, SpaceObjectPtr>(main_it->first, *local_it);
}

/*
pair<int,SpaceObjectPtr> SpaceIterator::operator->()
{
  return pair<int,SpaceObjectPtr>(main_it->first, *local_it);
}
*/

SpaceIterator& SpaceIterator::operator++()
{
  local_it++;
  while(local_it == main_it->second->end())
    {
      if((++main_it) == end)
	{
	  return *this;
	}
      local_it = main_it->second->begin();
    }
  return *this;
}

SpaceIterator SpaceIterator::operator++(int)
{
  SpaceIterator copy(*this);
  ++(*this);
  return copy;
}

bool SpaceIterator::operator!=(const SpaceIterator& other) const
{
  if(main_it == other.main_it)
    {
      if(main_it != end)
	{
	  return local_it != other.local_it;
	}
      return false;
    }
  return true;
}

bool SpaceIterator::operator==(const SpaceIterator& other) const
{
  if(main_it == other.main_it)
    {
      if(main_it != end)
	{
	  return local_it == other.local_it;
	}
      return true;
    }
  return false;
}

SpaceIterator& SpaceIterator::operator=(const SpaceIterator& copy)
{
  main_it = copy.main_it;
  local_it = copy.local_it;
  end = copy.end;
  return *this;
}

SpaceMap::SpaceMap(int w, int h, QWidget * parent, const char * name, WFlags f)
  : QWidget(parent, name, f)
  , baseSize(42)
  , m_width(w)
  , m_height(h)
  , rectCase(QPoint(0,0), QSize(baseSize+2,baseSize+2))
  , isSmall(false)
  , objRef( NULL )
{
  initGraphic();
  setMouseTracking(true);
  initVars();
}

int SpaceMap::caseId(QPoint p) const
{
  if((p.y() < 0) || (p.x() < 0) ||
     (p.y() >= m_height) || (p.x() >= m_width))
    {
      return -1;
    }
  return p.y()*m_width + p.x();
}

QPoint SpaceMap::casePos(int id) const
{
  int x = id%m_width;
  return QPoint(x, (id-x)/m_width);
}

QRect SpaceMap::caseRect(int id) const
{
  QPoint p(casePos(id));
  return caseRect(p);
}

QRect SpaceMap::caseRect(QPoint p) const
{
  return QRect(p*(baseSize+2), ((p*(baseSize+2))+QPoint(baseSize+1,baseSize+1)));
}

void SpaceMap::paintEvent(QPaintEvent *event)
{
  QRect rect(event->rect().topLeft(),event->rect().size());
  bitBlt(this, event->rect().topLeft(), &buffer, rect, CopyROP);
  if(rect.intersects(rectCase))
    {
      QPainter paint(this);
      paint.setPen(ONMOUSE);
      paint.drawRect(rectCase);
    }
}

void SpaceMap::mouseMoveEvent(QMouseEvent *event)
{
  int newCase = caseId(event->pos()/(baseSize+2));
  if(newCase != currentCase)
    {
      currentCase = newCase;
      if(pressed)
	{
	  tool->apply(newCase, *this);
	}
      QRect newRect(caseRect(newCase));
      QRect modif(newRect.unite(rectCase));
      rectCase = newRect;
      update(modif);
    }
}

void SpaceMap::mousePressEvent(QMouseEvent *event)
{
  curPos = event->pos();
  if(event->button() == LeftButton)
    {
      if(!tool.isNull())
        {
          pressedCase = caseId(event->pos()/(baseSize+2));
          if(tool->apply(pressedCase, *this))
	    {
	      pressed = true;
	    }
        }
    }
  else if(event->button() == RightButton)
    {
      currentCase = caseId(event->pos()/(baseSize+2));
      SpaceCasePtr c = getCase(currentCase);
      SpaceCase::iterator it;
      if(!c->empty())
	{
	  QPopupMenu *menu = new QPopupMenu(this);
	  for(it = c->begin() ; it != c->end() ; it++)
	    {
	      menu->insertItem((*it)->getName(), this, SLOT(DoNothing()), 0, (int)(*it).toPtr());
	    }
	  connect(menu, SIGNAL(activated(int)), this, SLOT(itemParameter(int)));
	  menu->popup(mapToGlobal(event->pos()));
	}
    }
}

void SpaceMap::mouseReleaseEvent(QMouseEvent * /* event */)
{
  if(pressed)
    {
      pressed = false;
    }
}

void SpaceMap::itemParameter(int id)
{
  SpaceObjectPtr object((SpaceObject*)id);
  object->changeParameters(this);
  getCase(currentCase)->repaint();
  update(caseRect(currentCase));
}

SpaceCasePtr SpaceMap::getCase(int pos)
{
  SpaceCasePtr& result = smap[pos];
  if(result.isNull())
    {
      result = new SpaceCase(pos, this);
    }
  return result;
}

QPixmap* SpaceMap::getPixmap()
{
  return &buffer;
}

void SpaceMap::setTool(SpaceToolPtr t)
{
  tool = t;
}

void SpaceMap::deleteObject(int id)
{
  curCase->deleteItem(id);
}

void SpaceMap::setCase(SpaceCasePtr c)
{
  curCase = c;
}

SpaceIterator SpaceMap::begin() const
{
  return SpaceIterator(smap);
}

SpaceIterator SpaceMap::end() const
{
  return SpaceIterator(smap,false);
}

int SpaceMap::width() const
{
  return m_width;
}

int SpaceMap::height() const
{
  return m_height;
}

SpaceMap::case_iterator SpaceMap::case_begin()
{
  return smap.begin();
}

SpaceMap::case_iterator SpaceMap::case_end()
{
  return smap.end();
}

SpaceMap::case_const_iterator SpaceMap::case_begin() const
{
  return smap.begin();
}

SpaceMap::case_const_iterator SpaceMap::case_end() const
{
  return smap.end();
}

void SpaceMap::reset(int w, int h)
{
  m_width = w;
  m_height = h;
  smap.clear();
  initGraphic();
  initVars( false );
  update();
}

void SpaceMap::initGraphic()
{
  int x,y;
  buffer = QPixmap((baseSize+2)*m_width,(baseSize+2)*m_height);
  QPainter paint;
  paint.begin(&buffer);
  paint.setPen(UNSELECT);
  paint.fillRect(0,0,(baseSize+2)*m_width,(baseSize+2)*m_height,BACKGROUND);
  for(x = 0 ; x < m_width ; x++)
    {
      paint.drawLine((baseSize+2)*x, 0, (baseSize+2)*x, (baseSize+2)*m_height);
      paint.drawLine((baseSize+2)*x+(baseSize+1), 0, (baseSize+2)*x+(baseSize+1), (baseSize+2)*m_height);
    }
  for(y = 0 ; y < m_height ; y++)
    {
      paint.drawLine(0, (baseSize+2)*y, (baseSize+2)*m_width, (baseSize+2)*y);
      paint.drawLine(0, (baseSize+2)*y+(baseSize+1), (baseSize+2)*m_width, (baseSize+2)*y+(baseSize+1));
    }
  QSize s((baseSize+2)*m_width, (baseSize+2)*m_height);
  setMinimumSize(s);
  setMaximumSize(s);
  setBaseSize(s);
  paint.end();
  case_iterator it_case;
  for( it_case = case_begin() ; it_case != case_end() ; ++it_case )
    {
      it_case->second->repaint();
    }
}

void SpaceMap::initVars( bool initTool )
{
  currentCase = -1;
  pressed = false;
  if( initTool )
    {
      tool = NULL;
    }
  curCase = NULL;
}

void SpaceMap::setSmall(bool small)
{
  isSmall = small;
  if(small)
    {
      baseSize = 21;
    }
  else
    {
      baseSize = 42;
    }
  initGraphic();
}

void SpaceMap::toggleSize()
{
  setSmall(isSmall^true);
}

void SpaceMap::setRef( ObjectRef* oref )
{
  objRef = oref;
}

