/***************************************************************************
                          spaceobject.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spaceobject.h"
#include "global.h"
#include "doorconfig.h"
#include "entryconfig.h"
#include "bulkconfig.h"
#include "exitconfig.h"
#include "extensibleconfig.h"

#include <qwmatrix.h>
#include <qdom.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qpushbutton.h>
#include <qspinbox.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qlabel.h>

#include <iostream>
#include <string>
#include <exception>
#include <typeinfo>

using namespace std;

SpaceObject::SpaceObject(ObjectRef* oref, int l, const QString n)
  : level(l)
  , name(n)
  , objRef( oref )
{ }

SpaceObject::SpaceObject(const SpaceObject& copy)
  : RefCountObject(copy)
  , level(copy.level)
  , name(copy.name)
  , objRef( copy.objRef )
{ }

SpaceObject::~SpaceObject()
{
}

int SpaceObject::getLevel()
{
  return level;
}

const QString& SpaceObject::getName()
{
  return name;
}

bool operator==(const SpaceObject& p1, const SpaceObject& p2)
{
  return p1.isIncluded(p2) && p2.isIncluded(p1);
}

bool operator<(const SpaceObject& o1, const SpaceObject& o2)
{
  return o1.LessThan(o2);
}

void SpaceObject::changeParameters(QWidget * /* parent */)
{ }

QPixmap *FloorObject::image = NULL;
QPixmap *FloorObject::icon = NULL;

FloorObject::FloorObject(ObjectRef* oref)
  : SpaceObject(oref, 1, "Floor")
{
  if(image == NULL)
    {
      image = new QPixmap(getThemeDir() + "case_42.png");
    }
  if(icon == NULL)
    {
      icon = new QPixmap(getThemeDir() + "case_21.png");
    }
}

FloorObject::FloorObject(const FloorObject& copy)
  : SpaceObject(copy)
{ }

const QPixmap& FloorObject::getImage(bool small)
{
  if(small)
    {
      return *icon;
    }
  return *image;
}

bool FloorObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      dynamic_cast<const FloorObject&>(object);
      return true;
    }
  catch(bad_cast&)
    {
      return false;
    }
}

QDomElement FloorObject::getXmlElement(QDomDocument& doc)
{
  return doc.createElement("floor");
}

void FloorObject::setFromXml(QDomElement& /*xml*/)
{
}

size_t FloorObject::hash() const
{
  return 1;// << 24;
}

bool FloorObject::LessThan(const SpaceObject& other) const
{
  try
    {
      dynamic_cast<const FloorObject&>(other);
      return false;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

SpaceObjectPtr FloorObject::getCopy()
{
  return new FloorObject(*this);
}

QPixmap *DoorObject::vert_image = NULL;
QPixmap *DoorObject::vert_icon = NULL;
QPixmap *DoorObject::hor_image = NULL;
QPixmap *DoorObject::hor_icon = NULL;

DoorObject::DoorObject(ObjectRef* oref, bool is_vertical)
  : SpaceObject(oref, 2)
  , vertical(is_vertical)
{
  if(vert_image == NULL)
    {
      hor_image = new QPixmap(getThemeDir() + "doorclose_h_42.png");
      hor_icon = new QPixmap(getThemeDir() + "doorclose_h_21.png");
      QWMatrix trans;
      trans.rotate(90);
      vert_image = new QPixmap(hor_image->xForm(trans));
      vert_icon = new QPixmap(hor_icon->xForm(trans));
    }
  setName();
}

DoorObject::DoorObject(const DoorObject& copy)
  : SpaceObject(copy)
  , vertical(copy.vertical)
{ }

SpaceObjectPtr DoorObject::getCopy()
{
  return new DoorObject(*this);
}

const QPixmap& DoorObject::getImage(bool small)
{
  if(small)
    {
      if(vertical)
	{
	  return *vert_icon;
	}
      return *hor_icon;
    }
  if(vertical)
    {
      return *vert_image;
    }
  return *hor_image;
}

bool DoorObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      const DoorObject& comp = dynamic_cast<const DoorObject&>(object);
      return vertical == comp.vertical;
    }
  catch(bad_cast&)
    {
      return false;
    }
}

QDomElement DoorObject::getXmlElement(QDomDocument& doc)
{
  QDomElement xml = doc.createElement("door"); 
  xml.setAttribute("vertical", (vertical?"yes":"no"));
  return xml;
}

void DoorObject::setFromXml(QDomElement& xml)
{
  QString vert = xml.attribute("vertical", "no");
  vertical = vert == "yes";
  setName();
}

size_t DoorObject::hash() const
{
  return 2;// << 24 + ((vertical?1:0)<<8);
}

bool DoorObject::LessThan(const SpaceObject& other) const
{
  try
    {
      const DoorObject& o = dynamic_cast<const DoorObject&>(other);
      return vertical < o.vertical;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

void DoorObject::changeParameters(QWidget *parent)
{
  DoorConfig config(parent, "Door Vertical Dlg", true);
  config.Vertical->setOn(vertical);
  config.Horizontal->setOn(!vertical);
  if(config.exec() == QDialog::Accepted)
    {
      vertical = config.Vertical->isOn();
      setName();
    }
}

void DoorObject::setName()
{
  if(vertical)
    {
      name = "Door vertical";
    }
  else
    {
      name = "Door horizontal";
    }
}

QPixmap *EntryObject::left_image = NULL;
QPixmap *EntryObject::right_image = NULL;
QPixmap *EntryObject::up_image = NULL;
QPixmap *EntryObject::down_image = NULL;
QPixmap *EntryObject::left_icon = NULL;
QPixmap *EntryObject::right_icon = NULL;
QPixmap *EntryObject::up_icon = NULL;
QPixmap *EntryObject::down_icon = NULL;

EntryObject::EntryObject(ObjectRef* oref, POSITION pos)
  : SpaceObject(oref, 1)
  , position(pos)
  , nb_blips(3)
{
  if(down_image == NULL)
    {
      QWMatrix trans;
      down_image = new QPixmap(getThemeDir() + "entryzone_42.png");
      down_icon = new QPixmap(getThemeDir() + "entryzone_21.png");
      trans.rotate(90);
      left_image = new QPixmap(down_image->xForm(trans));
      up_image = new QPixmap(left_image->xForm(trans));
      right_image = new QPixmap(up_image->xForm(trans));
      left_icon = new QPixmap(down_icon->xForm(trans));
      up_icon = new QPixmap(left_icon->xForm(trans));
      right_icon = new QPixmap(up_icon->xForm(trans));
    }
    setName();
}

EntryObject::EntryObject(const EntryObject& copy)
  : SpaceObject(copy)
  , position(copy.position)
  , nb_blips(copy.nb_blips)
{ }

SpaceObjectPtr EntryObject::getCopy()
{
  return new EntryObject(*this);
}

const QPixmap& EntryObject::getImage(bool small)
{
  switch(position)
    {
    case LEFT:
      if(small)
	return *left_icon;
      return *left_image;
    case RIGHT:
      if(small)
	return *right_icon;
      return *right_image;
    case DOWN:
      if(small)
	return *down_icon;
      return *down_image;
    case UP:
      if(small)
	return *up_icon;
      return *up_image;
    default:
      cerr << "Bad value for entry direction" << endl;
      if(small)
	return *up_icon;
      return *up_image;
    }
  throw std::string("ERROR");
}

bool EntryObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      const EntryObject& comp = dynamic_cast<const EntryObject&>(object);
      return position == comp.position;
    }
  catch(bad_cast&)
    {
      return false;
    }
}

void EntryObject::changeParameters(QWidget *parent)
{
  EntryConfig conf(parent, "Entry Direction Dlg", true);
  switch(position)
    {
    case LEFT:
      conf.Left->setOn(true);
      break;
    case RIGHT:
      conf.Right->setOn(true);
      break;
    case UP:
      conf.Up->setOn(true);
      break;
    case DOWN:
      conf.Down->setOn(true);
      break;
    default:
      cerr << "Bad value for entry direction" << endl;
      break;
    }
  conf.MaxBlip->setValue(nb_blips);
  if(conf.exec() == QDialog::Accepted)
    {
      if(conf.Left->isOn())
	{
	  position = LEFT;
	}
      else if(conf.Right->isOn())
	{
	  position = RIGHT;
	}
      else if(conf.Up->isOn())
	{
	  position = UP;
	}
      else
	{
	  position = DOWN;
	}
      nb_blips = conf.MaxBlip->text().toInt();
      setName();
    }
}

QDomElement EntryObject::getXmlElement(QDomDocument& doc)
{
  QDomElement xml = doc.createElement("entry"); 
  switch(position)
    {
    case LEFT:
      xml.setAttribute("direction", "east");
      break;
    case RIGHT:
      xml.setAttribute("direction", "west");
      break;
    case UP:
      xml.setAttribute("direction", "south");
      break;
    case DOWN:
      xml.setAttribute("direction", "north");
      break;
    default:
      cerr << "Bad value for entry direction" << endl;
      break;
    }
  xml.setAttribute("maxblip", QString::number(nb_blips));
  return xml;
}

void EntryObject::setFromXml(QDomElement& xml)
{
  QString pos = xml.attribute("direction");
  if(pos.isNull())
    {
      throw ObjectException();
    }
  if(pos == "east")
    {
      position = LEFT;
    }
  else if(pos == "west")
    {
      position = RIGHT;
    }
  else if(pos == "south")
    {
      position = UP;
    }
  else
    {
      position = DOWN;
    }
  QString max = xml.attribute("maxblip", "3");
  nb_blips = max.toInt();
  setName();
}

size_t EntryObject::hash() const
{
  return 3;// << 24 + (position << 8);
}

bool EntryObject::LessThan(const SpaceObject& other) const
{
  try
    {
      const EntryObject& o = dynamic_cast<const EntryObject&>(other);
      return position < o.position;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

void EntryObject::setName()
{
  switch(position)
    {
    case UP:
      name = "Entry zone up";
      break;
    case DOWN:
      name = "Entry zone down";
      break;
    case LEFT:
      name = "Entry zone left";
      break;
    case RIGHT:
      name = "Entry zone right";
      break;
    default:
      cerr << "Bad value for entry direction" << endl;
      break;
    }
  name += ", max blips : " + QString::number(nb_blips);
}

QPixmap *BulkHeadObject::vert_close_image = NULL;
QPixmap *BulkHeadObject::vert_close_icon = NULL;
QPixmap *BulkHeadObject::hor_close_image = NULL;
QPixmap *BulkHeadObject::hor_close_icon = NULL;
QPixmap *BulkHeadObject::vert_open_image = NULL;
QPixmap *BulkHeadObject::vert_open_icon = NULL;
QPixmap *BulkHeadObject::hor_open_image = NULL;
QPixmap *BulkHeadObject::hor_open_icon = NULL;

BulkHeadObject::BulkHeadObject(ObjectRef* oref, bool is_vertical)
  : SpaceObject(oref, 2)
  , vertical(is_vertical)
  , open(true)
  , can_open(false)
  , can_closed(true)
{
  if(vert_open_image == NULL)
    {
      hor_close_image = new QPixmap(getThemeDir() + "bulkclose_42.png");
      hor_close_icon = new QPixmap(getThemeDir() + "bulkclose_21.png");
      QWMatrix trans;
      trans.rotate(90);
      vert_close_image = new QPixmap(hor_close_image->xForm(trans));
      vert_close_icon = new QPixmap(hor_close_icon->xForm(trans));
      hor_open_image = new QPixmap(getThemeDir() + "bulkopen_42.png");
      hor_open_icon = new QPixmap(getThemeDir() + "bulkopen_21.png");
      vert_open_image = new QPixmap(hor_open_image->xForm(trans));
      vert_open_icon = new QPixmap(hor_open_icon->xForm(trans));
    }
  setName();
}

BulkHeadObject::BulkHeadObject(const BulkHeadObject& copy)
  : SpaceObject(copy)
  , vertical(copy.vertical)
  , open(copy.open)
  , can_open(copy.can_open)
  , can_closed(copy.can_closed)
{ }

SpaceObjectPtr BulkHeadObject::getCopy()
{
  return new BulkHeadObject(*this);
}

const QPixmap& BulkHeadObject::getImage(bool small)
{
  if(small)
    {
      if(vertical)
	{
	  if(open)
	    {
	      return *vert_open_icon;
	    }
	  return *vert_close_icon;
	}
      if(open)
	{
	  return *hor_open_icon;
	}
      return *hor_close_icon;
    }
  if(vertical)
    {
      if(open)
	{
	  return *vert_open_image;
	}
      return *vert_close_image;
    }
  if(open)
    {
      return *hor_open_image;
    }
  return *hor_close_image;
}

bool BulkHeadObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      const BulkHeadObject& comp = dynamic_cast<const BulkHeadObject&>(object);
      return vertical == comp.vertical;
    }
  catch(bad_cast&)
    {
      return false;
    }
}

void BulkHeadObject::changeParameters(QWidget *parent)
{
  BulkConfig conf(parent, "Bulkhead Vertical Dlg", true);
  conf.open->setChecked(open);
  conf.can_open->setChecked(can_open);
  conf.can_closed->setChecked(can_closed);
  if(vertical)
    {
      conf.vertical->setChecked(true);
    }
  else
    {
      conf.horizontal->setChecked(true);
    }
  if(conf.exec() == QDialog::Accepted)
    {
      open = conf.open->isChecked();
      can_open = conf.can_open->isChecked();
      can_closed = conf.can_closed->isChecked();
      vertical = conf.vertical->isChecked();
      setName();
    }
}

QDomElement BulkHeadObject::getXmlElement(QDomDocument& doc)
{
  QDomElement xml = doc.createElement("bulkhead"); 
  xml.setAttribute("open", (open?"yes":"no"));
  xml.setAttribute("can_open", (can_open?"yes":"no"));
  xml.setAttribute("can_closed", (can_closed?"yes":"no"));
  xml.setAttribute("vertical", (vertical?"yes":"no"));
  return xml;
}

void BulkHeadObject::setFromXml(QDomElement& xml)
{
  open = (xml.attribute("open", "yes") == "yes");
  can_open = (xml.attribute("can_open", "no") == "yes");
  can_closed = (xml.attribute("can_closed", "yes") == "yes");
  vertical = (xml.attribute("vertical", "no") == "yes");
  setName();
}

size_t BulkHeadObject::hash() const
{
  return 4;/* << 24 + ((vertical   ?1:0)<<8)
	         + ((open       ?1:0)<<9)
	         + ((can_open   ?1:0)<<10)
	         + ((can_closed ?1:0)<<11);*/
}

bool BulkHeadObject::LessThan(const SpaceObject& other) const
{
  try
    {
      const BulkHeadObject& o = dynamic_cast<const BulkHeadObject&>(other);
      if(vertical != o.vertical)
        {
          return vertical < o.vertical;
        }
      if(open != o.open)
        {
          return open < o.open;
        }
      if(can_open != o.can_open)
        {
          return can_open < o.can_open;
        }
      return can_closed < o.can_closed;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

void BulkHeadObject::setName()
{
  if(open)
    {
      name = "Opened";
    }
  else
    {
      name = "Closed";
    }
  if(can_open)
    {
      name += " openable";
    }
  if(can_closed)
    {
      name += " closable";
    }
  if(vertical)
    {
      name += " vertical";
    }
  else
    {
      name += " horizontal";
    }
  name += " bulkhead";
}

QPixmap *ExitObject::left_image = NULL;
QPixmap *ExitObject::right_image = NULL;
QPixmap *ExitObject::up_image = NULL;
QPixmap *ExitObject::down_image = NULL;
QPixmap *ExitObject::left_icon = NULL;
QPixmap *ExitObject::right_icon = NULL;
QPixmap *ExitObject::up_icon = NULL;
QPixmap *ExitObject::down_icon = NULL;

ExitObject::ExitObject(ObjectRef* oref, POSITION pos)
  : SpaceObject(oref, 1)
  , position(pos)
  , marine(true)
{
  if(down_image == NULL)
    {
      QWMatrix trans;
      down_image = new QPixmap(getThemeDir() + "exit_42.png");
      down_icon = new QPixmap(getThemeDir() + "exit_21.png");
      trans.rotate(90);
      left_image = new QPixmap(down_image->xForm(trans));
      up_image = new QPixmap(left_image->xForm(trans));
      right_image = new QPixmap(up_image->xForm(trans));
      left_icon = new QPixmap(down_icon->xForm(trans));
      up_icon = new QPixmap(left_icon->xForm(trans));
      right_icon = new QPixmap(up_icon->xForm(trans));
    }
  setName();
}

ExitObject::ExitObject(const ExitObject& copy)
  : SpaceObject(copy)
  , position(copy.position)
  , marine(copy.marine)
{ }

SpaceObjectPtr ExitObject::getCopy()
{
  return new ExitObject(*this);
}

const QPixmap& ExitObject::getImage(bool small)
{
  switch(position)
    {
    case LEFT:
      if(small)
	return *left_icon;
      return *left_image;
    case RIGHT:
      if(small)
	return *right_icon;
      return *right_image;
    case DOWN:
      if(small)
	return *down_icon;
      return *down_image;
    case UP:
      if(small)
	return *up_icon;
      return *up_image;
    default:
      cerr << "Bad value for exit direction" << endl;
      if( small )
        return *up_icon;
      return *up_image;
    }
  throw std::string("ERROR");
}

bool ExitObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      const ExitObject& comp = dynamic_cast<const ExitObject&>(object);
      return (position == comp.position) && (marine == comp.marine);
    }
  catch(bad_cast&)
    {
      return false;
    }
}

void ExitObject::changeParameters(QWidget *parent)
{
  ExitConfig conf(parent, "Exit config dlg", true);
  conf.marine->setChecked(marine);
  conf.genestealer->setChecked(!marine);
  switch(position)
    {
    case LEFT:
      conf.Left->setOn(true);
      break;
    case RIGHT:
      conf.Right->setOn(true);
      break;
    case UP:
      conf.Up->setOn(true);
      break;
    case DOWN:
      conf.Down->setOn(true);
      break;
    default:
      cerr << "Bad value for exit direction" << endl;
      break;
    }
  if(conf.exec() == QDialog::Accepted)
    {
      if(conf.Left->isOn())
	{
	  position = LEFT;
	}
      else if(conf.Right->isOn())
	{
	  position = RIGHT;
	}
      else if(conf.Up->isOn())
	{
	  position = UP;
	}
      else
	{
	  position = DOWN;
	}
      marine = conf.marine->isChecked();
      setName();
    }
}

QDomElement ExitObject::getXmlElement(QDomDocument& doc)
{
  QDomElement xml = doc.createElement("exit"); 
  if(marine)
    {
      xml.setAttribute("player", "marine");
    }
  else
    {
      xml.setAttribute("player", "genestealer");
    }
  switch(position)
    {
    case LEFT:
      xml.setAttribute("direction", "east");
      break;
    case RIGHT:
      xml.setAttribute("direction", "west");
      break;
    case UP:
      xml.setAttribute("direction", "south");
      break;
    case DOWN:
      xml.setAttribute("direction", "north");
      break;
    default:
      cerr << "Bad value for exit direction" << endl;
      break;
    }
  return xml;
}

void ExitObject::setFromXml(QDomElement& xml)
{
  QString player = xml.attribute("player");
  QString dir = xml.attribute("direction");
  if(player.isNull() || dir.isNull())
    {
      throw ObjectException();
    }
  marine = (player == "marine");
  if(dir == "north")
    {
      position = DOWN;
    }
  else if(dir == "south")
    {
      position = UP;
    }
  else if(dir == "east")
    {
      position = LEFT;
    }
  else
    {
      position = RIGHT;
    }
  setName();
}

size_t ExitObject::hash() const
{
  return 5;// << 24 + (position << 8) + (marine?1:0)<<10;
}

bool ExitObject::LessThan(const SpaceObject& other) const
{
  try
    {
      const ExitObject& o = dynamic_cast<const ExitObject&>(other);
      if(marine != o.marine)
        {
          return marine < o.marine;
        }
      return position < o.position;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

void ExitObject::setName()
{
  switch(position)
    {
    case UP:
      name = "Exit zone up";
      break;
    case DOWN:
      name = "Exit zone down";
      break;
    case LEFT:
      name = "Exit zone left";
      break;
    case RIGHT:
      name = "Exit zone right";
      break;
    default:
      cerr << "Bad value for exit direction" << endl;
      break;
    }
  if(marine)
    {
      name += " for marine";
    }
  else
    {
      name += " for genestealer";
    }
}

QPixmap *ExtensibleObject::toxinduct_image = NULL;
QPixmap *ExtensibleObject::archives_image = NULL;
QPixmap *ExtensibleObject::machinery_image = NULL;
QPixmap *ExtensibleObject::toxinduct_icon = NULL;
QPixmap *ExtensibleObject::archives_icon = NULL;
QPixmap *ExtensibleObject::machinery_icon = NULL;

ExtensibleObject::ExtensibleObject(ObjectRef* oref)
  : SpaceObject(oref, 2)
  , desc("Toxin duct")
  , graphic(TOXINDUCT)
  , empty(false)
  , attackable(false)
  , fireable(false)
  , shootwith(true)
  , opendoorwith(true)
  , shootable(0)
  , takeable(0)
  , removeable(0)
  , transferable(0)
  , obj_id( Referencer::null_reference )
{
  if(toxinduct_image == NULL)
    {
      toxinduct_image = new QPixmap(getThemeDir() + "toxinduct_42.png");
      toxinduct_icon = new QPixmap(getThemeDir() + "toxinduct_21.png");
      archives_image = new QPixmap(getThemeDir() + "archives_42.png");
      archives_icon = new QPixmap(getThemeDir() + "archives_21.png");
      machinery_image = new QPixmap(getThemeDir() + "machinery_42.png");
      machinery_icon = new QPixmap(getThemeDir() + "machinery_21.png");
    }
  setName();
}

ExtensibleObject::ExtensibleObject(const ExtensibleObject& copy)
  : SpaceObject(copy)
  , desc(copy.desc)
  , graphic(copy.graphic)
  , empty(copy.empty)
  , attackable(copy.attackable)
  , fireable(copy.fireable)
  , shootwith(copy.shootwith)
  , opendoorwith(copy.opendoorwith)
  , shootable(copy.shootable)
  , takeable(copy.takeable)
  , removeable(copy.removeable)
  , transferable(copy.transferable)
  , obj_id( Referencer::null_reference )
{ }

ExtensibleObject::~ExtensibleObject()
{
  if( obj_id != Referencer::null_reference )
    {
      try
        {
          objRef->eraseReference( obj_id );
          obj_id = Referencer::null_reference;
        }
      catch( ... )
        {
          cerr << "Unexpected exception in ~ExtensibleObject" << endl;
        }
    }
}

SpaceObjectPtr ExtensibleObject::getCopy()
{
  return new ExtensibleObject(*this);
}

void ExtensibleObject::setName()
{
  QString obj_name = "";
  if( obj_id != Referencer::null_reference )
    {
      obj_name = objRef->getName( obj_id ) + " ";
    }
  name = "Extensible " + obj_name + ": " + desc;
}

bool ExtensibleObject::isIncluded(const SpaceObject& object) const
{
  try
    {
      const ExtensibleObject& comp = dynamic_cast<const ExtensibleObject&>(object);
      bool result = true;
      result &= (desc == comp.desc);
      result &= (graphic == comp.graphic);
      result &= (empty == comp.empty);
      result &= (attackable == comp.attackable);
      result &= (fireable == comp.fireable);
      result &= (shootwith == comp.shootwith);
      result &= (opendoorwith == comp.opendoorwith);
      result &= (attackable == comp.attackable);
      result &= (takeable == comp.takeable);
      result &= (removeable == comp.removeable);
      result &= (transferable == comp.transferable);
      return result;
    }
  catch(bad_cast&)
    {
      return false;
    }
}

void ExtensibleObject::changeParameters(QWidget *parent)
{
  ExtensibleConfig conf(parent, "Extensible config dlg", true);
  conf.ref->hide();
  conf.label_ref->hide();
  conf.obj_name->setText( objRef->getName( obj_id ) );
  switch(graphic)
    {
    case TOXINDUCT:
      conf.toxinduct->setChecked(true);
      break;
    case ARCHIVES:
      conf.archives->setChecked(true);
      break;
    case MACHINERY:
      conf.machinery->setChecked(true);
      break;
    }
  conf.empty->setChecked(empty);
  conf.attackable->setChecked(attackable);
  conf.fireable->setChecked(fireable);
  conf.shootwith->setChecked(shootwith);
  conf.opendoorwith->setChecked(opendoorwith);
  conf.shoot_PA->setValue(shootable);
  conf.take_PA->setValue(takeable);
  conf.remove_PA->setValue(removeable);
  conf.transfer_PA->setValue(transferable);
  conf.desc->setText(desc);
  if(conf.exec() == QDialog::Accepted)
    {
      if(conf.toxinduct->isChecked())
	{
	  graphic = TOXINDUCT;
	}
      else if(conf.archives->isChecked())
	{
	  graphic = ARCHIVES;
	}
      else
	{
	  graphic = MACHINERY;
	}
      empty = conf.empty->isChecked();
      attackable = conf.attackable->isChecked();
      fireable = conf.fireable->isChecked();
      shootwith = conf.shootwith->isChecked();
      opendoorwith = conf.opendoorwith->isChecked();
      shootable = conf.shoot_PA->text().toInt();
      takeable = conf.take_PA->text().toInt();
      removeable = conf.remove_PA->text().toInt();
      transferable = conf.transfer_PA->text().toInt();
      desc = conf.desc->text();
      QString obj_name = conf.obj_name->text();
      if( obj_name != objRef->getName( obj_id ) )
        {
          if( obj_name != "" )
            {
              if( obj_id == Referencer::null_reference )
                {
                  obj_id = objRef->refAsObject( obj_name, this );
                }
              else
                {
                  objRef->renameRef( obj_id, obj_name );
                }
            }
          else
            {
              objRef->eraseReference( obj_id );
              obj_id = Referencer::null_reference;
            }
        }
      setName();
    }
}

size_t ExtensibleObject::hash() const
{
  return 6;/* << 24 + (graphic << 8)
	         + (empty << 10)
	         + (attackable << 11)
	         + (fireable << 12)
	         + (shootwith << 13)
	         + (opendoorwith << 14)
	         + ((shootable + takeable + removeable + transferable) << 15);*/
}

bool ExtensibleObject::LessThan(const SpaceObject& other) const
{
  try
    {
      const ExtensibleObject& o = dynamic_cast<const ExtensibleObject&>(other);
      if(graphic != o.graphic)
        {
          return graphic < o.graphic;
        }
      if(desc != o.desc )
        {
          return desc< o.desc;
        }
      if(empty != o.empty )
        {
          return empty< o.empty;
        }
      if(attackable != o.attackable  )
        {
          return attackable < o.attackable ;
        }
      if(fireable != o.fireable  )
        {
          return fireable < o.fireable ;
        }
      if(shootwith != o.shootwith  )
        {
          return shootwith < o.shootwith ;
        }
      if(opendoorwith != o.opendoorwith  )
        {
          return opendoorwith < o.opendoorwith ;
        }
      if(shootable != o.shootable  )
        {
          return shootable < o.shootable ;
        }
      if(takeable != o.takeable  )
        {
          return takeable < o.takeable ;
        }
      if(removeable != o.removeable  )
        {
          return removeable < o.removeable ;
        }
      return transferable < o.transferable ;
    }
  catch(bad_cast&)
    {
      return hash() < other.hash();
    }
}

const QPixmap& ExtensibleObject::getImage(bool small)
{
  if(small)
    {
      switch(graphic)
	{
	case TOXINDUCT:
	  return *toxinduct_icon;
	case ARCHIVES:
	  return *archives_icon;
	case MACHINERY:
	  return *machinery_icon;
	}
    }
  switch(graphic)
    {
    case TOXINDUCT:
      return *toxinduct_image;
    case ARCHIVES:
      return *archives_image;
    case MACHINERY:
      return *machinery_image;
    }
  return *machinery_image;
}

QDomElement ExtensibleObject::getXmlElement(QDomDocument& doc)
{
  QDomElement xml = doc.createElement("extensible"); 
  xml.setAttribute("desc", desc);
  switch(graphic)
    {
    case TOXINDUCT:
      xml.setAttribute("graphic", "toxinduct");
      break;
    case ARCHIVES:
      xml.setAttribute("graphic", "archives");
      break;
    case MACHINERY:
      xml.setAttribute("graphic", "machinery");
      break;
    }
  xml.setAttribute("empty", (empty?"yes":"no"));
  xml.setAttribute("attackable", (attackable?"yes":"no"));
  xml.setAttribute("shootable",QString::number(shootable));
  xml.setAttribute("fireable",(fireable?"yes":"no"));
  xml.setAttribute("takeable",QString::number(takeable));
  xml.setAttribute("removeable", QString::number(removeable));
  xml.setAttribute("transferable", QString::number(transferable));
  xml.setAttribute("shootwith",(shootwith?"yes":"no"));
  xml.setAttribute("usedoorwith",(opendoorwith?"yes":"no"));
  return xml;
}

void ExtensibleObject::setFromXml(QDomElement& xml)
{
  QString d = xml.attribute("desc");
  QString grap = xml.attribute("graphic");
  if(d.isNull() || grap.isNull())
    {
      throw ObjectException();
    }
  desc = d;
  if(grap == "toxinduct")
    {
      graphic = TOXINDUCT;
    }
  else if(grap == "archives")
    {
      graphic = ARCHIVES;
    }
  else
    {
      graphic = MACHINERY;
    }
  empty = (xml.attribute("empty", "no") == "yes");
  attackable = (xml.attribute("attackable", "no") == "yes");
  shootable = xml.attribute("shootable", "0").toInt();
  fireable = (xml.attribute("fireable", "no") == "yes");
  takeable = xml.attribute("takeable", "0").toInt();
  removeable = xml.attribute("removeable", "0").toInt();
  transferable = xml.attribute("transferable", "0").toInt();
  shootwith = (xml.attribute("shootwith", "yes") == "yes");
  opendoorwith = (xml.attribute("usedoorwith", "yes") == "yes");
  QString obj_name = xml.attribute( "name", "" );
  if( obj_name != "" )
    {
      obj_id = objRef->refAsObject( obj_name, this );
    }
  setName();
}

void ExtensibleObject::breakReferences()
{
  if( obj_id != Referencer::null_reference )
    {
      objRef->eraseReference( obj_id );
      obj_id = Referencer::null_reference;
    }
}

