/***************************************************************************
                          spaceobject.h  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _MAPOBJECT_HH_
#define _MAPOBJECT_HH_

#include "config.h"
#include "rcobject.h"
#include "global.h"
#include "objectref.h"

#include <qpixmap.h>
#include <list>

#include <map>

using namespace std;

DECLARE_FWD_PTR(SpaceObject);

class QDomElement;
class QDomDocument;

class ObjectException
{
};

class SpaceObject : public RefCountObject
{
public:
  SpaceObject(ObjectRef* oref, int level, const QString n = "");
  SpaceObject(const SpaceObject& copy);
  virtual ~SpaceObject();
  virtual const QPixmap& getImage(bool small = false) = 0;
  virtual const QString& getName();
  // Return true if the object in parameter is a subtype of the
  // current object and that every variables of the current
  // class are equal (Rem: this is NOT equallity)
  virtual bool isIncluded(const SpaceObject& object) const = 0;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc) = 0;
  virtual void setFromXml(QDomElement& xml) = 0;
  virtual size_t hash() const = 0;
  virtual SpaceObjectPtr getCopy() = 0;
  int getLevel();
  virtual bool LessThan(const SpaceObject& other) const = 0;
  virtual void breakReferences() {}
protected:
  int level;
  QString name;
  ObjectRef *objRef;
};

bool operator<(const SpaceObject& o1, const SpaceObject& o2);

bool operator==(const SpaceObject& p1, const SpaceObject& p2);

class FloorObject : public SpaceObject
{
public:
  FloorObject(ObjectRef* oref);
  FloorObject(const FloorObject& copy);
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
protected:
  static QPixmap *image, *icon;
};

class DoorObject : public SpaceObject
{
public:
  DoorObject(ObjectRef* oref, bool is_vertical = true);
  DoorObject(const DoorObject& copy);
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
protected:
  void setName();
  bool vertical;
  static QPixmap *vert_image, *vert_icon, *hor_image, *hor_icon;
};

class EntryObject : public SpaceObject
{
public:
  EntryObject(ObjectRef* oref, POSITION pos = DOWN);
  EntryObject(const EntryObject& copy);
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
protected:
  void setName();
  POSITION position;
  int nb_blips;
  static QPixmap *left_image, *right_image, *up_image, *down_image;
  static QPixmap *left_icon, *right_icon, *up_icon, *down_icon;
};

class BulkHeadObject : public SpaceObject
{
public:
  BulkHeadObject(ObjectRef* oref, bool is_vertical = true);
  BulkHeadObject(const BulkHeadObject& copy);
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
protected:
  void setName();
  bool vertical, open, can_open, can_closed;
  static QPixmap *vert_open_image, *vert_open_icon, *hor_open_image, *hor_open_icon;
  static QPixmap *vert_close_image, *vert_close_icon, *hor_close_image, *hor_close_icon;
};

class ExitObject : public SpaceObject
{
public:
  ExitObject(ObjectRef* oref, POSITION pos = DOWN);
  ExitObject(const ExitObject& copy);
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
protected:
  void setName();
  POSITION position;
  bool marine;
  static QPixmap *left_image, *right_image, *up_image, *down_image;
  static QPixmap *left_icon, *right_icon, *up_icon, *down_icon;
};

class ExtensibleObject : public SpaceObject
{
public:
  enum GRAPHIC
  {
    TOXINDUCT,
    ARCHIVES,
    MACHINERY
  };

  ExtensibleObject(ObjectRef* oref);
  ExtensibleObject(const ExtensibleObject& copy);
  ~ExtensibleObject();
  virtual const QPixmap& getImage(bool small = false);
  virtual bool isIncluded(const SpaceObject& object) const;
  virtual void changeParameters(QWidget *parent);
  virtual QDomElement getXmlElement(QDomDocument& doc);
  virtual size_t hash() const;
  virtual SpaceObjectPtr getCopy();
  virtual void setFromXml(QDomElement& xml);
  virtual bool LessThan(const SpaceObject& other) const;
  virtual void breakReferences();
protected:
  void setName();
  QString desc;
  GRAPHIC graphic;
  bool empty, attackable, fireable, shootwith, opendoorwith;
  int shootable, takeable, removeable, transferable;
  static QPixmap *toxinduct_image, *archives_image, *machinery_image;
  static QPixmap *toxinduct_icon, *archives_icon, *machinery_icon;
  Referencer::reference_id obj_id;
};

#endif
