/***************************************************************************
                          spaceref.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spaceref.h"

#include <iostream>

void Referencer::bind( Referencer::reference_id ref )
{
  refCenter->bind( ref );
}

void Referencer::unbind( Referencer::reference_id ref )
{
  refCenter->unbind( ref );
}

QString Referencer::getName( Referencer::reference_id ref ) const
{
  return refCenter->getName( ref );
}

Referencer::reference_id Referencer::getId( const QString& name ) const
{
  return refCenter->getId( name );
}

QString Referencer::getName_nofail( Referencer::reference_id ref ) const
{
  return refCenter->getName_nofail( ref );
}

Referencer::reference_id Referencer::getId_nofail( const QString& name )
{
  return refCenter->getId_nofail( name );
}

void Referencer::renameRef( Referencer::reference_id ref, const QString& new_name )
{
  refCenter->renameRef( ref, new_name );
}

const Referencer::reference_id Referencer::null_reference = -1;

ReferenceCenter::ReferenceCenter()
  : current_id( 0 )
{
}

Referencer::reference_id ReferenceCenter::createId()
{
  return current_id++;
}

Referencer::reference_id ReferenceCenter::createRef(const QString& name, Referencer *r )
{
  Referencer::reference_id id;
  if( exists( name ) )
    {
      throw NameConflictException( name );
    }
  map<QString, Referencer::reference_id>::iterator found = failed_string.find( name );
  if( found != failed_string.end() )
    {
      id = found->second;
      existing_string[ name ] = found->second;
      failed_string.erase( found );
      references_type::iterator ref_it = failsafe.find( id );
      ref_it->second.referencer = r;
      references[ id ] = ref_it->second;
      failsafe.erase( ref_it );
    }
  else
    {
      id = createId();
      references[id] = Reference( name, r );
      existing_string[ name ] = id;
    }
  return id;
}

void ReferenceCenter::bind( Referencer::reference_id ref )
{
  if( ref == Referencer::null_reference )
    return;
  if( !exists( ref ) )
    {
      throw InexistingRefException( ref );
    }
  references[ref].bound++;
}

void ReferenceCenter::unbind( Referencer::reference_id ref )
{
  if( ref == Referencer::null_reference )
    return;
  if( !exists( ref ) )
    {
      throw InexistingRefException( ref );
    }
  references[ref].bound--;
}

bool ReferenceCenter::exists( Referencer::reference_id ref ) const
{
  return (references.find( ref ) != references.end());
}

bool ReferenceCenter::exists(const QString& name) const
{
  return (existing_string.find(name) != existing_string.end());
}

QString ReferenceCenter::createName() const
{
  QString base = "unnamed";
  QString result;
  int i = 1;
  do
    {
      result = base + QString::number(i++);
    } while(existing_string.find(result) != existing_string.end());
  return result;
}

int ReferenceCenter::referenced( Referencer::reference_id ref ) const
{
  references_type::const_iterator it =  references.find( ref );
  if(it == references.end())
    {
      return 0;
    }
  return it->second.bound;
}

void ReferenceCenter::clear()
{
  references.clear();
  existing_string.clear();
}

void ReferenceCenter::eraseRef( Referencer::reference_id ref )
{
  references_type::iterator found = references.find( ref );
  if( found == references.end() )
    {
      throw InexistingRefException( ref );
    }
  existing_string.erase( found->second.name );
  references.erase( found );
}

QString ReferenceCenter::getName( Referencer::reference_id ref ) const
{
  if( ref == Referencer::null_reference )
    return "";
  references_type::const_iterator found = references.find( ref );
  if( found == references.end() )
    {
      throw InexistingRefException( ref );
    }
  return found->second.name;
}

Referencer::reference_id ReferenceCenter::getId( const QString& name ) const
{
  if( name == "" )
    return Referencer::null_reference;
  map<QString, Referencer::reference_id>::const_iterator found = existing_string.find( name );
  if( found == existing_string.end() )
    {
      throw InexistingRefException( name );
    }
  return found->second;
}

void ReferenceCenter::renameRef( Referencer::reference_id ref, const QString& newname )
{
  references_type::iterator found = references.find( ref );
  if( found == references.end() )
    {
      throw InexistingRefException( ref );
    }
  if( existing_string.find( newname ) != existing_string.end() )
    {
      throw NameConflictException( newname );
    }
  existing_string.erase( found->second.name );
  existing_string[ newname ] = ref;
  found->second.name = newname;
}

Referencer::reference_id ReferenceCenter::getId_nofail( const QString& name )
{
  map<QString, Referencer::reference_id>::const_iterator found = existing_string.find( name );
  if( found == existing_string.end() )
    {
      found = failed_string.find( name );
      if( found == failed_string.end() )
        {
          Referencer::reference_id ref = createId();
          failed_string[ name ] = ref;
          failsafe[ ref ] = Reference( name );
          return ref;
        }
    }
  return found->second;
}

QString ReferenceCenter::getName_nofail( Referencer::reference_id ref ) const
{
  if( ref == Referencer::null_reference )
    return "";
  references_type::const_iterator found = references.find( ref );
  if( found == references.end() )
    {
      found = failsafe.find( ref );
      if( found == failsafe.end() )
        {
          throw InexistingRefException( ref );
        }
    }
  return found->second.name;
}

void ReferenceCenter::bind_nofail( Referencer::reference_id ref )
{
  if( ref == Referencer::null_reference )
    return;
  references_type::iterator found = references.find( ref );
  if( found == references.end() )
    {
      found = failsafe.find( ref );
      if( found == failsafe.end() )
        {
          throw InexistingRefException( ref );
        }
    }
  found->second.bound++;
}

bool ReferenceCenter::exists_nofail( Referencer::reference_id ref )
{
  if( ref == Referencer::null_reference )
    return true;
  references_type::iterator found = references.find( ref );
  if( found == references.end() )
    {
      found = failsafe.find( ref );
      if( found == failsafe.end() )
        {
          return false;
        }
    }
  return true;

}

bool ReferenceCenter::exists_nofail( const QString& name )
{
  if( name == "" )
    return true;
  map<QString, Referencer::reference_id>::const_iterator found = existing_string.find( name );
  if( found == existing_string.end() )
    {
      found = failed_string.find( name );
      if( found == failed_string.end() )
        {
          return false;
        }
    }
  return true;
}

bool ReferenceCenter::checkFailsafe()
{
  if( !failed_string.empty() )
  {
    map<QString,Referencer::reference_id>::iterator it;
    for( it = failed_string.begin() ; it != failed_string.end() ; ++it )
      {
        cerr << it->first << ",  ";
      }
    cerr << endl;
  }
  return failed_string.empty();
}

