/***************************************************************************
                          spacezones.cpp  -  description
                             -------------------
    copyright            : (C) 2003 by Pierre Barbier de Reuille
    email                : harbort@nerim.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "spacezones.h"
#include "functions.h"
#include "global.h"
#include "errordlg.h"

#include <algorithm>
#include <iostream>

using namespace std;

#include "myzonesetedit.h"

#include <qlineedit.h>
#include <qpoint.h>

saveCase::saveCase(QDomElement& r, const SpaceMap& m, QDomDocument& d)
  : root(r)
  , smap(m)
  , doc(d)
{ }

void saveCase::operator()(int pos)
{
  QPoint pt = smap.casePos(pos);
  QDomElement case_xml = doc.createElement("case");
  case_xml.setAttribute("x", pt.x());
  case_xml.setAttribute("y", pt.y());
  root.appendChild(case_xml);
}

saveRef::saveRef(const QString& t, QDomElement& r, QDomDocument& d)
  : root(r)
  , doc(d)
  , type(t)
{ }

void saveRef::operator()(const QString& ref_name)
{
  QDomElement xml = doc.createElement(type);
  QDomElement ref = doc.createElement("ref");
  ref.setAttribute("name", ref_name);
  xml.appendChild(ref);
  root.appendChild(xml);
}

SpaceZones::SpaceZones(const SpaceZones& copy)
  : RefCountObject(copy)
  , Referencer(copy)
  , zones(copy.zones)
  , zone_sets(copy.zone_sets)
  , smap(copy.smap)
  , doc(NULL)
{ }

SpaceZones::SpaceZones(ReferenceCenter* rc)
  : Referencer(rc)
  , smap(NULL)
  , doc(NULL)
{ }

Referencer::reference_id SpaceZones::setZone(const QString& zone_name, const set<int>& zone)
{
  Referencer::reference_id ref;
  if( refCenter->exists( zone_name ) )
    {
      ref = refCenter->getId( zone_name );
      if( zones.find( ref ) != zones.end() )
        {
          zones[ ref ] = zone;
        }
    }
  ref = refCenter->createRef(zone_name, this);
  zones[ ref ] = zone;
  return ref;
}

Referencer::reference_id SpaceZones::setZone(const QDomElement& xml)
{
  set<int> result;
  QString name = xml.attribute("name");
  if(!name.isNull())
    {
      if(refCenter->exists( name ))
        {
          cout << "Error, this name \"" << name << "\" is already in used" << endl;
          ErrorNameConflict dlg( name );
          dlg.exec();
          return Referencer::null_reference;
        }
    }
  QDomElement xml_cases = xml.firstChild().toElement();
  loadCases lc(*smap, *this);
  if(xml_cases.tagName() == "ref")
    {
      QString ref_name = xml_cases.attribute("name");
      if(ref_name.isNull())
        {
          cerr << "Error, ref must have name attribute" << endl;
          ErrorXmlSyntax dlg;
          dlg.exec();
          return Referencer::null_reference;
        }
      Referencer::reference_id ref_zone = refCenter->getId_nofail( ref_name );
      zones_type::iterator it = zones.find( ref_zone );
      if( ( it == zones.end() ) || name.isNull() )
        {
          return ref_zone;
        }
      result = it->second;
    }
  else
    {
      while(!xml_cases.isNull())
        {
          set<int> local = lc(xml_cases);
#ifdef _WIN32
          for (set<int>::iterator i = local.begin(); i!=local.end(); ++i)
            {
              result.insert(*i);
            }
#else
          result.insert(local.begin(), local.end());
#endif
          xml_cases = xml_cases.nextSibling().toElement();
        }
    }
  Referencer::reference_id ref;
  if(name.isNull())
    {
      name = refCenter->createName();
    }
  ref = refCenter->createRef(name, this);
  zones[ ref ] = result;
  return ref;
}

set<int> SpaceZones::getZone( Referencer::reference_id ref )
{
  zones_type::iterator it = zones.find( ref );
  if(it == zones.end())
    {
      cerr << "Error, inexisting zone reference" << endl;
      throw InexistingRefException( ref );
    }
  return it->second;
}

Referencer::reference_id SpaceZones::setZone(const QString& zone_name)
{
  Referencer::reference_id ref = Referencer::null_reference;
  if(refCenter->exists(zone_name))
    {
      ref = refCenter->getId( zone_name );
      if( zones.find( ref ) == zones.end() )
        {
          cerr << "Error, name " << zone_name << " conflict with another object" << endl;
          throw NameConflictException(zone_name);
        }
    }
  set<int> zone;
  SpaceMap::case_iterator it_case;
  for( it_case = smap->case_begin() ; it_case != smap->case_end() ; ++it_case )
    {
      if( it_case->second->selected() )
        {
          zone.insert( it_case->first );
        }
    }
  set<int>::iterator it;
  for( it = zone.begin() ; it != zone.end() ; ++it )
    {
      smap->getCase( *it )->toggle_select();
    }
  if(zone.empty())
    {
      throw InvalidDataException();
    }
  if( ref == Referencer::null_reference )
    {
      ref = refCenter->createRef(zone_name, this );
    }
  zones[ ref ] = zone;
  return ref;
}

Referencer::reference_id SpaceZones::setZoneSet(const QString& zone_set_name, const set<Referencer::reference_id>& zone_set)
{
  Referencer::reference_id ref = refCenter->createRef(zone_set_name, this );
  set<Referencer::reference_id>::const_iterator it_zs;
  for( it_zs = zone_set.begin() ; it_zs != zone_set.end() ; ++it_zs )
    {
      refCenter->bind( *it_zs );
    }
  zone_sets[ ref ] = zone_set;
  return ref;
}

Referencer::reference_id SpaceZones::setZoneSet(const QDomElement& xml)
{
  set<Referencer::reference_id> result;
  QString name = xml.attribute("name");
  if(!name.isNull())
    {
      if(refCenter->exists(name))
        {
          cout << "Error, this name \"" << name << "\" is already in used" << endl;
          ErrorNameConflict dlg( name );
          dlg.exec();
          return Referencer::null_reference;
        }
    }
  QDomElement xml_zone = xml.firstChild().toElement();
  if(xml_zone.tagName() == "ref")
    {
      QString ref_name = xml_zone.attribute("name");
      Referencer::reference_id zone_ref = refCenter->getId_nofail( ref_name );
      zone_sets_type::iterator it = zone_sets.find(zone_ref);
      if( ( it == zone_sets.end() ) || name.isNull() )
        {
          return zone_ref;
        }
      result = it->second;
    }
  else
    {
      while(!xml_zone.isNull())
        {
          if(xml_zone.tagName() != "zone")
            {
              cerr << "Error, zone set must have zones or ref as child only" << endl;
              throw InvalidXmlException();
            }
          Referencer::reference_id ref_zone = setZone(xml_zone);
          refCenter->bind_nofail( ref_zone );
          result.insert( ref_zone );
          xml_zone = xml_zone.nextSibling().toElement();
        }
    }
  Referencer::reference_id ref = Referencer::null_reference;
  if(name.isNull())
    {
      name = refCenter->createName();
    }
  ref = refCenter->createRef(name, this);
  zone_sets[ ref ] = result;
  return ref;
}

list<set<int> > SpaceZones::getZoneSet( Referencer::reference_id ref )
{
  zone_sets_type::iterator it = zone_sets.find( ref );
  if(it == zone_sets.end())
    {
      cerr << "Inexisting zone set reference" << endl;
      throw InexistingRefException( ref );
    }
  list<set<int> > result;
  set<Referencer::reference_id>::iterator it_zone;
  for(it_zone = it->second.begin() ; it_zone != it->second.end() ; it_zone++)
    {
      result.push_back(getZone(*it_zone));
    }
  return result;
}

Referencer::reference_id SpaceZones::setZoneSet(const QString& zone_set_name)
{
  set<QString> existing;
  set<QString> used;
  zones_type::iterator it_z;
  for( it_z = zones.begin() ; it_z != zones.end() ; ++it_z )
    {
      existing.insert( refCenter->getName( it_z->first ) );
    }
  Referencer::reference_id ref = Referencer::null_reference;
  if( refCenter->exists( zone_set_name ) )
    {
      ref = refCenter->getId( zone_set_name );
      zone_sets_type::iterator it = zone_sets.find( ref );
      set<Referencer::reference_id>::const_iterator it_zr;
      for( it_zr = it->second.begin() ; it_zr != it->second.end() ; ++it_zr )
        {
          used.insert( refCenter->getName( *it_zr ) );
        }
    }
  MyZoneSetEdit dlg(zone_set_name, existing, used, smap, "Zone set edition", true);
  if(dlg.exec() == QDialog::Accepted)
    {
      QString name = dlg.set_name->text();
      if(name == "")
        {
          cerr << "Error, empty name" << endl;
          throw InvalidDataException();
        }
      if( name != zone_set_name )
        {
          if( refCenter->exists(name) )
            {
              cerr << "Error, name " << name << " conflicts with another object" << endl;
              throw NameConflictException(name);
            }
          else if( ref != Referencer::null_reference )
            {
              refCenter->renameRef( ref, name );
            }
          else
            {
              ref = refCenter->createRef(name, this);
            }
        }
      else if( ref == Referencer::null_reference )
        {
          ref = refCenter->createRef( name, this );
        }
      set<QString>::iterator it;
      set<Referencer::reference_id> &zs = zone_sets[ ref ];
      zs.clear();
      for( it = dlg.used.begin() ; it != dlg.used.end() ; ++it )
        {
          refCenter->bind( refCenter->getId( *it ) );
          zs.insert( refCenter->getId( *it ) );
        }
      if(dlg.used.empty())
        {
          cerr << "No zone selected " << endl;
          throw InvalidDataException();
        }
    }
  return 0;
}

QDomElement SpaceZones::getXmlElement( Referencer::reference_id ref ) const
{
  zones_type::const_iterator zone_it = zones.find( ref );
  if(zone_it != zones.end())
    {
      QString name = refCenter->getName( ref );
      QDomElement xml = doc->createElement("zone");
      xml.setAttribute("name", name);
      set<int>::const_iterator it;
      saveCase sc( xml, *smap, *doc );
      for( it = zone_it->second.begin() ; it != zone_it->second.end() ; ++it )
        {
          sc( *it );
        }
      return xml;
    }
  zone_sets_type::const_iterator zs_it = zone_sets.find( ref );
  if(zs_it != zone_sets.end())
    {
      QDomElement xml = doc->createElement("zone_set");
      QString name = refCenter->getName( ref );
      xml.setAttribute("name", name);
      saveRef sr( "zone", xml, *doc );
      set<Referencer::reference_id>::const_iterator it;
      for( it = zs_it->second.begin() ; it != zs_it->second.end() ; ++it )
        {
          sr( refCenter->getName( *it ) );
        }
      return xml;
    }
  return QDomElement();
}

void SpaceZones::eraseReference( Referencer::reference_id ref )
{
  if( ref == Referencer::null_reference )
    return;
  QString name = refCenter->getName( ref );
  if( refCenter->referenced( ref ) )
    {
      cerr << "Error, object " << name << " still in use" << endl;
      throw StillRefException( refCenter->getName( ref ) );
    }
  zones_type::iterator zone_it = zones.find( ref );
  zone_sets_type::iterator zone_set_it = zone_sets.find( ref );
  if( ( zone_it == zones.end()) && ( zone_set_it == zone_sets.end() ) )
    {
      throw InexistingRefException(name);
    }
  if(zone_it != zones.end())
    {
      zones.erase( zone_it );
    }
  else
    {
      set<Referencer::reference_id>::iterator it;
      for( it = zone_set_it->second.begin() ; it != zone_set_it->second.end() ; ++it )
        {
          refCenter->unbind( *it );
        }
      zone_sets.erase(zone_set_it);
    }
  refCenter->eraseRef(ref);
}

void SpaceZones::setDoc(QDomDocument* d)
{
  doc = d;
}

void SpaceZones::setMap(SpaceMap* m)
{
  smap = m;
}

void SpaceZones::clear()
{
  zones.clear();
  zone_sets.clear();
}

loadCases::loadCases(const SpaceMap& m, SpaceZones& z)
  : smap(m)
  , zones(z)
{ }

set<int> loadCases::operator()(QDomElement& root)
{
  set<int> result;
  if(root.tagName() == "case")
    {
      int x,y;
      QString sx, sy;
      sx = root.attribute("x");
      sy = root.attribute("y");
      if(sx.isNull() || sy.isNull())
        {
          cerr << "Error, case element must have x and y attributes" << endl;
          ErrorXmlSyntax dlg;
          dlg.exec();
        }
      else
        {
          x = sx.toInt();
          y = sy.toInt();
          result.insert(smap.caseId(QPoint(x,y)));
        }
    }
  else if(root.tagName() == "rect")
    {
      int x1,y1,x2,y2;
      QString sx1, sx2, sy1, sy2;
      sx1 = root.attribute("x1");
      sy1 = root.attribute("y1");
      sx2 = root.attribute("x2");
      sy2 = root.attribute("y2");
      if(sx1.isNull() || sy1.isNull() || sx2.isNull() || sy2.isNull())
        {
          cerr << "Error, rect element must have x1,y1,x2,y2 attributes" << endl;
          ErrorXmlSyntax dlg;
          dlg.exec();
          return result;
        }
      x1 = sx1.toInt();
      y1 = sy1.toInt();
      x2 = sx2.toInt();
      y2 = sy2.toInt();
      for(int x = x1 ; x <= x2 ; x++)
        {
          for(int y = y1 ; y <= y2 ; y++)
            {
              result.insert(smap.caseId(QPoint(x,y)));
            }
        }
    }
  else if(root.tagName() == "zone_set")
    {
      Referencer::reference_id sz_ref = zones.setZoneSet(root);
      if( sz_ref == Referencer::null_reference )
        {
          return result;
        }
      set<Referencer::reference_id>& l = zones.zone_sets[ sz_ref ];
      set<Referencer::reference_id>::iterator it;
      for(it = l.begin() ; it != l.end() ; it++)
        {
          set<int>& hs = zones.zones[*it];
#ifdef _WIN32
          for (set<int>::iterator win_it = hs.begin(); win_it!=hs.end(); ++win_it)
            {
              result.insert(*win_it);
            }
#else
          result.insert(hs.begin(), hs.end());
#endif
        }
    }
  else if(root.tagName() == "ref")
    {
      QString name = root.attribute("name");
      if(name.isNull())
        {
          cerr << "Error, ref must have a name" << endl;
          ErrorXmlSyntax dlg;
          dlg.exec();
          return result;
        }
      Referencer::reference_id ref = zones.getId_nofail( name );
      SpaceZones::zones_type::iterator it = zones.zones.find( ref );
      if(it != zones.zones.end())
        {
          return it->second;
        }
    }
  return result;
}

void SpaceZones::clean()
{
  zone_sets_type::iterator it_zs;
  zones_type::iterator it_z;
  list<Referencer::reference_id> to_erase;
  for( it_zs = zone_sets.begin() ; it_zs != zone_sets.end() ; ++it_zs )
    {
      Referencer::reference_id ref = it_zs->first;
      if( refCenter->referenced( ref ) == 0 )
        {
          QString name = refCenter->getName( ref );
          if( name.find( "unnamed" ) == 0 )
            {
              to_erase.push_back( ref );
            }
        }
    }
  list<Referencer::reference_id>::iterator it;
  for( it = to_erase.begin() ; it != to_erase.end() ; ++it )
    {
      eraseReference( *it );
    }
  to_erase.clear();
  for( it_z = zones.begin() ; it_z != zones.end() ; ++it_z )
    {
      Referencer::reference_id ref = it_z->first;
      if( refCenter->referenced( ref ) == 0 )
        {
          QString name = refCenter->getName( ref );
          if( name.find( "unnamed" ) == 0 )
            {
              to_erase.push_back( ref );
            }
        }
    }
  for( it = to_erase.begin() ; it != to_erase.end() ; ++it )
    {
      eraseReference( *it );
    }
}

